# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import copy
import datetime
from operator import attrgetter

import faker as faker_
import pytest
from freezegun import freeze_time
from nark.items.activity import Activity
from nark.items.category import Category
from nark.items.fact import Fact
from nark.items.tag import Tag

faker = faker_.Faker()


class TestCategory(object):
    @classmethod
    def as_repr(cls, category):
        if category is None:
            return repr(category)
        repr_f = "Category(_name={name}, deleted={deleted}, hidden={hidden}, pk={pk})"
        repred = repr_f.format(
            pk=repr(category.pk),
            name=repr(category.name),
            deleted=repr(category.deleted),
            hidden=repr(category.hidden),
        )
        return repred

    def test_init_valid(self, name_string_valid_parametrized, pk_valid_parametrized):
        """Make sure that Category constructor accepts all valid values."""
        category = Category(name_string_valid_parametrized, pk_valid_parametrized)
        assert category.name == name_string_valid_parametrized
        assert category.pk == pk_valid_parametrized

    def test_init_invalid(self, name_string_invalid_parametrized):
        """Make sure that Category constructor rejects all invalid values."""
        with pytest.raises(ValueError):
            Category(name_string_invalid_parametrized)

    def test_as_tuple_include_pk(self, category):
        """
        Make sure categories tuple representation works as intended and pk
        is included.
        """
        deleted = False
        hidden = False
        assert category.as_tuple() == (category.pk, category.name, deleted, hidden)

    def test_as_tuple_exclude_pf(self, category):
        """
        Make sure categories tuple representation works as intended and pk
        is excluded.
        """
        deleted = False
        hidden = False
        cat_tuple = category.as_tuple(include_pk=False)
        our_tuple = (False, category.name, deleted, hidden)
        assert cat_tuple == our_tuple

    def test_equal_fields_true(self, category):
        """
        Make sure that two categories that differ only in their PK compare equal.
        """
        other_category = copy.deepcopy(category)
        other_category.pk = 1
        assert category.equal_fields(other_category)

    def test_equal_fields_false(self, category):
        """
        Make sure that two categories that differ not only in their PK
        compare unequal.
        """
        other_category = copy.deepcopy(category)
        other_category.pk = 1
        other_category.name += 'foobar'
        assert category.equal_fields(other_category) is False

    def test__eq__false(self, category):
        """Make sure that two distinct categories return ``False``."""
        other_category = copy.deepcopy(category)
        other_category.pk = 1
        assert category is not other_category
        assert category != other_category

    def test__eq__true(self, category):
        """Make sure that two identical categories return ``True``."""
        other_category = copy.deepcopy(category)
        assert category is not other_category
        assert category == other_category

    def test_is_hashable(self, category):
        """Test that ``Category`` instances are hashable."""
        assert hash(category)

    def test_hash_method(self, category):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(category) == hash(category.as_tuple())

    def test_hash_different_between_instances(self, category_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Category.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(category_factory()) != hash(category_factory())

    def test__str__(self, category):
        """Test string representation."""
        assert '{name}'.format(name=category.name) == str(category)

    def test__repr__(self, category):
        """Test representation method."""
        result = repr(category)
        assert isinstance(result, str)
        expectation = TestCategory.as_repr(category)
        assert result == expectation


class TestActivity(object):
    @classmethod
    def as_repr(cls, activity):
        if activity is None:
            return repr(activity)
        repred = (
            "Activity(_name={name}, category={category}, "
            "deleted={deleted}, hidden={hidden}, pk={pk})"
        ).format(
            pk=repr(activity.pk),
            name=repr(activity.name),
            category=TestCategory.as_repr(activity.category),
            deleted=repr(activity.deleted),
            hidden=repr(activity.hidden),
        )
        return repred

    def test_init_valid(
        self,
        name_string_valid_parametrized,
        pk_valid_parametrized,
        category_valid_parametrized,
        deleted_valid_parametrized,
    ):
        """Test that init accepts all valid values."""
        activity = Activity(
            name_string_valid_parametrized,
            pk=pk_valid_parametrized,
            category=category_valid_parametrized,
            deleted=deleted_valid_parametrized,
        )
        assert activity.name == name_string_valid_parametrized
        assert activity.pk == pk_valid_parametrized
        assert activity.category == category_valid_parametrized
        assert activity.deleted == bool(deleted_valid_parametrized)

    def test_init_invalid(self, name_string_invalid_parametrized):
        """
        Test that init rejects all invalid values.

        Note:
            Right now, the only aspect that can have concievable invalid value
            is the name.
        """
        with pytest.raises(ValueError):
            Activity(name_string_invalid_parametrized)

    def test_create_from_composite(self, activity):
        result = Activity.create_from_composite(activity.name, activity.category.name)
        assert result.name == activity.name
        assert result.category == activity.category

    def test_as_tuple_include_pk(self, activity):
        """Make sure that conversion to a tuple matches our expectations."""
        expecting = (
            activity.pk,
            activity.name,
            (
                activity.category.pk,
                activity.category.name,
                activity.category.deleted,
                activity.category.hidden,
            ),
            activity.deleted,
            activity.hidden,
        )
        assert activity.as_tuple() == expecting

    def test_as_tuple_exclude_pk(self, activity):
        """Make sure that conversion to a tuple matches our expectations."""
        expecting = (
            False,
            activity.name,
            (
                False,
                activity.category.name,
                activity.category.deleted,
                activity.category.hidden,
            ),
            activity.deleted,
            activity.hidden,
        )
        assert activity.as_tuple(include_pk=False) == expecting

    def test_equal_fields_true(self, activity):
        """
        Make sure that two activities that differ only in their PK compare equal.
        """
        other = copy.deepcopy(activity)
        other.pk = 1
        assert activity.equal_fields(other)

    def test_equal_fields_false(self, activity):
        """
        Make sure that two activities that differ not only in their PK
        compare unequal.
        """
        other = copy.deepcopy(activity)
        other.pk = 1
        other.name += 'foobar'
        assert activity.equal_fields(other) is False

    def test__eq__false(self, activity):
        """Make sure that two distinct activities return ``False``."""
        other = copy.deepcopy(activity)
        other.pk = 1
        assert activity is not other
        assert activity != other

    def test__eq__true(self, activity):
        """Make sure that two identical activities return ``True``."""
        other = copy.deepcopy(activity)
        assert activity is not other
        assert activity == other

    def test_is_hashable(self, activity):
        """Test that ``Category`` instances are hashable."""
        assert hash(activity)

    def test_hash_method(self, activity):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(activity) == hash(activity.as_tuple())

    def test_hash_different_between_instances(self, activity_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Category.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(activity_factory()) != hash(activity_factory())

    def test__str__without_category(self, activity):
        activity.category = None
        assert str(activity) == '{name}'.format(name=activity.name)

    def test__str__with_category(self, activity):
        assert str(activity) == '{name} ({category})'.format(
            name=activity.name, category=activity.category.name)

    def test__repr__with_category(self, activity):
        """Make sure our debugging representation matches our expectations."""
        result = repr(activity)
        assert isinstance(result, str)
        expectation = TestActivity.as_repr(activity)
        assert result == expectation

    def test__repr__without_category(self, activity):
        """Make sure our debugging representation matches our expectations."""
        activity.category = None
        result = repr(activity)
        assert isinstance(result, str)
        expectation = TestActivity.as_repr(activity)
        assert result == expectation


class TestTag(object):
    @classmethod
    def as_repr(cls, tag):
        if tag is None:
            return repr(tag)
        repred = "Tag(_name={name}, deleted={deleted}, hidden={hidden}, pk={pk})".format(
            pk=repr(tag.pk),
            name=repr(tag.name),
            deleted=repr(tag.deleted),
            hidden=repr(tag.hidden),
        )
        return repred

    def test_init_valid(self, name_string_valid_parametrized_tag, pk_valid_parametrized):
        """Make sure that Tag constructor accepts all valid values."""
        tag = Tag(name_string_valid_parametrized_tag, pk_valid_parametrized)
        assert tag.name == name_string_valid_parametrized_tag
        assert tag.pk == pk_valid_parametrized

    def test_init_invalid(self, name_string_invalid_parametrized_tag):
        """Make sure that Tag constructor rejects all invalid values."""
        with pytest.raises(ValueError):
            Tag(name_string_invalid_parametrized_tag)

    def test_as_tuple_include_pk(self, tag):
        """Make sure tags tuple representation works as intended and pk is included."""
        deleted = False
        hidden = False
        assert tag.as_tuple() == (tag.pk, tag.name, deleted, hidden)

    def test_as_tuple_exclude_pf(self, tag):
        """Make sure tags tuple representation works as intended and pk is excluded."""
        deleted = False
        hidden = False
        assert tag.as_tuple(include_pk=False) == (False, tag.name, deleted, hidden)

    def test_equal_fields_true(self, tag):
        """Make sure that two tags that differ only in their PK compare equal."""
        other_tag = copy.deepcopy(tag)
        other_tag.pk = 1
        assert tag.equal_fields(other_tag)

    def test_equal_fields_false(self, tag):
        """Make sure that two tags that differ not only in their PK compare unequal."""
        other_tag = copy.deepcopy(tag)
        other_tag.pk = 1
        other_tag.name += 'foobar'
        assert tag.equal_fields(other_tag) is False

    def test__eq__false(self, tag):
        """Make sure that two distinct tags return ``False``."""
        other_tag = copy.deepcopy(tag)
        other_tag.pk = 1
        assert tag is not other_tag
        assert tag != other_tag

    def test__eq__true(self, tag):
        """Make sure that two identical categories return ``True``."""
        other_tag = copy.deepcopy(tag)
        assert tag is not other_tag
        assert tag == other_tag

    def test_is_hashable(self, tag):
        """Test that ``Tag`` instances are hashable."""
        assert hash(tag)

    def test_hash_method(self, tag):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(tag) == hash(tag.as_tuple())

    def test_hash_different_between_instances(self, tag_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Tag.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(tag_factory()) != hash(tag_factory())

    def test__str__(self, tag):
        """Test string representation."""
        assert '{name}'.format(name=tag.name) == str(tag)

    def test__repr__(self, tag):
        """Test representation method."""
        result = repr(tag)
        assert isinstance(result, str)
        expectation = TestTag.as_repr(tag)
        assert result == expectation


class TestFact(object):
    def test_fact_init_valid(
        self,
        activity,
        start_end_datetimes,
        pk_valid_parametrized,
        description_valid_parametrized,
        tag_list_valid_parametrized,
    ):
        """Make sure valid values instaniate a Fact."""
        fact = Fact(
            activity,
            start_end_datetimes[0],
            start_end_datetimes[1],
            pk_valid_parametrized,
            description_valid_parametrized,
            tag_list_valid_parametrized,
        )
        assert fact.activity == activity
        assert fact.pk == pk_valid_parametrized
        assert fact.description == description_valid_parametrized
        assert fact.start == start_end_datetimes[0]
        assert fact.end == start_end_datetimes[1]
        # tag_list_valid_parametrized is a set() of name strings.
        names = list(tag_list_valid_parametrized)
        tags = set([Tag(pk=None, name=name) for name in names])
        tags = sorted(list(tags), key=attrgetter('name'))
        assert fact.tags_sorted == tags

    @pytest.mark.parametrize(
        ('factoid', 'time_hint', 'lenient', 'should_err'),
        [
            (
                '12:00 - 14:00 foo@bar, rumpelratz',
                'verify_both',
                False,
                None,
            ),
            (
                '12:00 - 14:00 foo',
                'verify_both',
                True,
                'Expected to find an Activity name.',
            ),
            (
                'foo@bar',
                'verify_none',
                True,
                None,
            ),
            (
                '12:00-14:00 foo@bar',
                'verify_both',
                False,
                None,
            ),
            # Test seconds (2018-08-16: they no longer get truncated).
            (
                '12:00:11 - 14:00:59 baz@bat',
                'verify_both',
                False,
                None,
            ),
            # Test just start and end, no activity, category, tags, nor description.
            (
                '12:00:11 - 13:01',
                'verify_both',
                True,
                'Expected to find an Activity name.',
            ),
            # Test just a start time.
            (
                '13:01:22',
                'verify_start',
                True,
                'Expected to find an Activity name.',
            ),
        ],
    )
    def test_create_from_factoid_valid(self, factoid, time_hint, lenient, should_err):
        """Make sure that a valid raw fact creates a proper Fact."""
        fact, err = Fact.create_from_factoid(
            factoid, time_hint=time_hint, lenient=lenient,
        )
        assert fact
        assert str(err) == str(should_err)

# FIXME: The invalid_raw_fact_parametrized fixture won't fail anymore,
#        because clock time is considered relative until explicitly closed --
#
# FIXME: We need to add fix_times to these nark tests!
#
#
#    def test_create_from_factoid_invalid(self, invalid_raw_fact_parametrized):
#        """Make sure invalid string raises an exception."""
#        with pytest.raises(ValueError):
#            fact, err = Fact.create_from_factoid(invalid_raw_fact_parametrized)

    @pytest.mark.parametrize(
        ('raw_fact', 'expectations'),
        [
            (
                '-7 foo@bar, palimpalum',
                {
                    'start': None,
                    'end': None,
                    'activity': '-7 foo',
                    'category': 'bar',
                    'description': 'palimpalum',
                },
            ),
            (
                # FIXME: (lb): This is not parsing correctly!
                #   The Activity name is being set to '-7: foo'
                #   Though maybe that's because no `time_hint`!
                '-7: foo@bar, palimpalum',
                {
                    'start': None,
                    'end': None,
                    'activity': '-7: foo',
                    'category': 'bar',
                    'description': 'palimpalum',
                },
            ),
        ]
    )
    @freeze_time('2015-05-02 18:07')
    def test_create_from_factoid_with_delta_no_time_hint(self, raw_fact, expectations):
        fact, err = Fact.create_from_factoid(raw_fact)
        assert fact.start == expectations['start']
        assert fact.end == expectations['end']
        assert fact.activity.name == expectations['activity']
        assert fact.activity.category.name == expectations['category']
        assert fact.description == expectations['description']
        assert not err

    @pytest.mark.parametrize(
        ('factoid', 'expectations'),
        [
            (
                '-7 foo@bar, palimpalum',
                {
                    # FIXME/MAYBE: Should we make an intermediate time parser
                    # that resolves relative times? Then we could expect:
                    #   'start': datetime.datetime(2015, 5, 2, 18, 0, 0),
                    'start': '-7',
                    'end': None,
                    'activity': 'foo',
                    'category': 'bar',
                    'description': 'palimpalum',
                },
            ),
        ],
    )
    @freeze_time('2015-05-02 18:07')
    def test_create_from_factoid_with_delta_time_hint_start(self, factoid, expectations):
        fact, err = Fact.create_from_factoid(factoid, time_hint='verify_start')
        assert fact.start == expectations['start']
        assert fact.end == expectations['end']
        assert fact.activity.name == expectations['activity']
        assert fact.activity.category.name == expectations['category']
        assert fact.description == expectations['description']
        assert not err

    @pytest.mark.parametrize(
        'start',
        [
            None,
            faker.date_time(),
            '10',
            '+10',
            '-10h5m',
        ],
    )
    def test_start_valid(self, fact, start):
        """Make sure that valid arguments get stored by the setter."""
        fact.start = start
        assert fact.start == start

    @pytest.mark.parametrize(
        'start',
        [
            faker.date_time().strftime('%y-%m-%d %H:%M'),  # string, not datetime
            'not relative',
            '+10d'  # Not supported
        ],
    )
    def test_start_invalid(self, fact, start):
        """Make sure that trying to store dateimes as strings throws an error."""
        with pytest.raises(TypeError):
            fact.start = start

    @pytest.mark.parametrize('end', [None, faker.date_time()])
    def test_end_valid(self, fact, end):
        """Make sure that valid arguments get stored by the setter."""
        fact.end = end
        assert fact.end == end

    def test_end_invalid(self, fact):
        """Make sure that trying to store dateimes as strings throws an error."""
        with pytest.raises(TypeError):
            fact.end = faker.date_time().strftime('%y-%m-%d %H:%M')

    def test_description_valid(self, fact, description_valid_parametrized):
        """Make sure that valid arguments get stored by the setter."""
        fact.description = description_valid_parametrized
        assert fact.description == description_valid_parametrized

    def test_delta(self, fact):
        """Make sure that valid arguments get stored by the setter."""
        assert fact.delta() == fact.end - fact.start

    @freeze_time('2015-05-02 18:07')
    def test_delta_no_end(self, fact):
        """Make sure that a missing end datetime results in ``delta=None``."""
        # See FactFactory for default start/end.
        fact.end = None
        # NOTE: With freezegun, both now() and utcnow() are the same.
        assert fact.delta() == (datetime.datetime.now() - fact.start)

    @pytest.mark.parametrize(
        'offset',
        [
            (
                15,
                {
                    '%M': '15',
                    '%H:%M': '00:15',
                    'HHhMMm': ' 0 hours 15 minutes',
                    '': '15.00 mins.',
                },
            ),
            (
                452,
                {
                    '%M': '452',
                    '%H:%M': '07:32',
                    'HHhMMm': ' 7 hours 32 minutes',
                    '': '7.53 hours',
                },
            ),
            (
                912,
                {
                    '%M': '912',
                    '%H:%M': '15:12',
                    'HHhMMm': '15 hours 12 minutes',
                    '': '15.20 hours',
                },
            ),
            (
                61,
                {
                    '%M': '61',
                    '%H:%M': '01:01',
                    'HHhMMm': ' 1 hour   1 minute ',
                    '': '1.02 hours',
                },
            ),
        ],
    )
    def test_format_delta_valid_style(
        self,
        fact,
        offset,
        start_end_datetimes_from_offset_now,
        string_delta_style_parametrized,
    ):
        """Make sure that the resulting string matches our expectation."""
        end_offset, expectation = offset
        fact.start, fact.end = start_end_datetimes_from_offset_now(end_offset)
        result = fact.format_delta(style=string_delta_style_parametrized)
        assert result == expectation[string_delta_style_parametrized]

    def test_format_delta_invalid_style(self, fact):
        """Ensure that passing an invalid format will raise an exception."""
        with pytest.raises(ValueError):
            fact.format_delta(style='foobar')

    def test_category_property(self, fact):
        """Make sure the property returns this facts category."""
        assert fact.category == fact.activity.category

    def test_serialized_string(self, fact):
        """
        Ensure a serialized string with full information matches our expectation.
        """
        expect_f = '{start} to {end} {activity}@{category}: #{tag}: {description}'
        expectation = expect_f.format(
            start=fact.start.strftime('%Y-%m-%d %H:%M:%S'),
            end=fact.end.strftime('%Y-%m-%d %H:%M:%S'),
            activity=fact.activity.name,
            category=fact.category.name,
            tag=sorted(list(fact.tags), key=attrgetter('name'))[0].name,
            description=fact.description
        )
        result = fact.get_serialized_string()
        assert isinstance(result, str)
        assert result == expectation

    @pytest.mark.parametrize(('values', 'expectation'), (
        (
            {
                'start': datetime.datetime(2016, 1, 1, 18),
                'end': datetime.datetime(2016, 1, 1, 19),
                'activity': Activity('homework', category=Category('school')),
                'tags': set([Tag('math'), Tag('science')]),
                'description': 'something clever ...',
            },
            '2016-01-01 18:00:00 to 2016-01-01 19:00:00 '
            'homework@school: #math #science: something clever ...',
        ),
        (
            {
                'start': datetime.datetime(2016, 1, 1, 18),
                'end': datetime.datetime(2016, 1, 1, 19),
                'activity': Activity('homework', category=None),
                'tags': set([Tag('math'), Tag('science'), Tag('science fiction')]),
                'description': 'something',
            },
            '2016-01-01 18:00:00 to 2016-01-01 19:00:00 '
            'homework@: #math #science #science fiction: something',
        ),
        (
            {
                'start': datetime.datetime(2016, 1, 1, 18),
                'end': datetime.datetime(2016, 1, 1, 19),
                'activity': Activity('homework', category=Category('school')),
                'tags': set(),
                'description': 'something clever ...',
            },
            '2016-01-01 18:00:00 to 2016-01-01 19:00:00 '
            'homework@school: something clever ...',
        ),
        (
            {
                'start': datetime.datetime(2016, 1, 1, 18),
                'end': datetime.datetime(2016, 1, 1, 19),
                'activity': Activity('homework', category=Category('school')),
                'tags': set([Tag('science'), Tag('math')]),
                'description': '',
            },
            '2016-01-01 18:00:00 to 2016-01-01 19:00:00 '
            'homework@school: #math #science',
        ),
        (
            {
                'start': datetime.datetime(2016, 1, 1, 18),
                'end': datetime.datetime(2016, 1, 1, 19),
                'activity': Activity('homework', category=Category('school')),
                'tags': set(),
                'description': '',
            },
            '2016-01-01 18:00:00 to 2016-01-01 19:00:00 '
            'homework@school',
        ),
        (
            {
                'start': None,
                'end': datetime.datetime(2016, 1, 1, 19),
                'activity': Activity('homework', category=Category('school')),
                'tags': set([Tag('math'), Tag('science')]),
                'description': 'something clever ...',
            },
            # FIXME/2018-08-17 12:25: Update factoid parse to recognize
            # 'to' and 'at' prefix to distinguish between verify_end, verify_start?
            # and then anything else is verify_both or verify_none?? hrmmmm...
            # maybe the answer is a 2nd parse-factoid wrapper, i.e.,
            #   one parser for verify_hint, and one parser for unknown-hint...
            '2016-01-01 19:00:00 homework@school: #math #science: something clever ...',
        ),
        (
            {
                'start': None,
                'end': None,
                'activity': Activity('homework', category=Category('school')),
                'tags': set([Tag('math'), Tag('science')]),
                'description': 'something clever ...',
            },
            # FIXME: Make new parse wrapper that checks for 'to' 'at', or date.
            #   Then look for 'to <date>', 'at <date>', etc.
            #   Fall back to what? Expect no dates? Both dates?
            #   Problem really is that I feel a Fact with no start or end
            #     is really an invalid Fact!! So this Factoid should be a
            #     problem, right?:
            'homework@school: #math #science: something clever ...',
        ),
        (
            {
                'start': datetime.datetime(2016, 1, 1, 18),
                'end': None,
                'activity': Activity('homework', category=Category('school')),
                'tags': set([Tag('math'), Tag('science')]),
                'description': 'something clever ...',
            },
            '2016-01-01 18:00:00 to <now> '
            'homework@school: #math #science: something clever ...',
        ),
    ))
    def test_serialized_string_various_missing_values(self, fact, values, expectation):
        """
        Make sure the serialized string is correct even if some information is missing.
        """
        for attribute, value in values.items():
            setattr(fact, attribute, value)
        assert fact.get_serialized_string() == expectation

    def test_as_tuple_include_pk(self, fact):
        """Make sure that conversion to a tuple matches our expectations."""
        assert fact.as_tuple() == (
            fact.pk,
            fact.activity.as_tuple(include_pk=True),
            fact.start,
            fact.end,
            fact.description,
            frozenset(fact.tags),
            fact.deleted,
            fact.split_from,
        )

    def test_as_tuple_exclude_pk(self, fact):
        """Make sure that conversion to a tuple matches our expectations."""
        assert fact.as_tuple(include_pk=False) == (
            False,
            fact.activity.as_tuple(include_pk=False),
            fact.start,
            fact.end,
            fact.description,
            frozenset([tag.as_tuple(include_pk=False) for tag in fact.tags]),
            fact.deleted,
            fact.split_from,
        )

    def test_equal_fields_true(self, fact):
        """Make sure that two facts that differ only in their PK compare equal."""
        other = copy.deepcopy(fact)
        other.pk = 1
        assert fact.equal_fields(other)

    def test_equal_fields_false(self, fact):
        """Make sure that two facts that differ not only in their PK compare unequal."""
        other = copy.deepcopy(fact)
        other.pk = 1
        other.description += 'foobar'
        assert fact.equal_fields(other) is False

    def test__eq__false(self, fact):
        """Make sure that two distinct facts return ``False``."""
        other = copy.deepcopy(fact)
        other.pk = 1
        assert fact is not other
        assert fact != other

    def test__eq__true(self, fact):
        """Make sure that two identical facts return ``True``."""
        other = copy.deepcopy(fact)
        assert fact is not other
        assert fact == other

    def test_is_hashable(self, fact):
        """Test that ``Fact`` instances are hashable."""
        assert hash(fact)

    def test_hash_method(self, fact):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(fact) == hash(fact.as_tuple())

    def test_hash_different_between_instances(self, fact_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Fact.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(fact_factory()) != hash(fact_factory())

    def test__str__(self, fact):
        expect_f = '{start} to {end} {activity}@{category}: {tags}: {description}'
        expectation = expect_f.format(
            start=fact.start.strftime('%Y-%m-%d %H:%M:%S'),
            end=fact.end.strftime('%Y-%m-%d %H:%M:%S'),
            activity=fact.activity.name,
            category=fact.category.name,
            tags=fact.tagnames(),
            description=fact.description,
        )
        assert str(fact) == expectation

    def test__str__no_end(self, fact):
        fact.end = None
        expect_f = "{start} to <now> {activity}@{category}: {tags}: {description}"
        expectation = expect_f.format(
            start=fact.start.strftime('%Y-%m-%d %H:%M:%S'),
            activity=fact.activity.name,
            category=fact.category.name,
            tags=fact.tagnames(),
            description=fact.description,
        )
        assert str(fact) == expectation

    def test__str__no_start_no_end(self, fact):
        fact.start = None
        fact.end = None
        expectation = '{activity}@{category}: {tags}: {description}'.format(
            activity=fact.activity.name,
            category=fact.category.name,
            tags=fact.tagnames(),
            description=fact.description,
        )
        assert str(fact) == expectation

    # (lb): It might be nice to do snapshot testing. However, that won't
    # work unless we disable the random string generator we use to make up
    # item names.
    #
    # Here's what a snapshot test might look like:
    #
    #   def test__repr__snapshot(self, snapshot, fact):
    #       """
    #       Test repr() against snapshot. Save time not writing test expectation.
    #       """
    #       result = repr(fact)
    #       snapshot.assert_match(result)
    #
    # In lieu of that, we re-generate the repr() herein. Which makes me
    # feel weird, like we're just duplicating what Fact.__repr__ already
    # does.

    def assert_fact_repr(self, the_fact):
        # (lb): I feel somewhat dirty with the test__repr__ methods.
        #   I feel like these should be snapshot tests, and not tests
        #   that require manual labor to maintain a tedious string
        #   builder that basically mimics the behavior of the methods
        #   that we're testing. Blech.
        tag_parts = []
        for tag in the_fact.tags:
            tag_parts.append(TestTag.as_repr(tag))
        tags = ', '.join(tag_parts)
        expect_f = (
            "Fact("
            "_description={description}, "
            "_end={end}, "
            "_start={start}, "
            "activity={activity}, "
            "deleted={deleted}, "
            "pk={pk}, "
            "split_from={split_from}, "
            "tags=[{tags}]"
            ")"
        )
        expectation = expect_f.format(
            pk=repr(the_fact.pk),
            split_from=repr(the_fact.split_from),
            start=repr(the_fact.start),
            end=repr(the_fact.end),
            activity=TestActivity.as_repr(the_fact.activity),
            tags=tags,
            description=repr(the_fact.description),
            deleted=repr(the_fact.deleted),
        )
        result = repr(the_fact)
        assert isinstance(result, str)
        assert result == expectation

    def test__repr__(self, fact):
        """Make sure our debugging representation matches our expectations."""
        self.assert_fact_repr(fact)

    def test__repr__no_end(self, fact):
        """Test that facts without end datetime are represented properly."""
        result = repr(fact)
        assert isinstance(result, str)
        fact.end = None
        self.assert_fact_repr(fact)

    def test__repr__no_start_no_end(self, fact):
        """Test that facts without times are represented properly."""
        fact.start = None
        fact.end = None
        self.assert_fact_repr(fact)

    def test__str__no_tags(self, fact):
        fact.tags = []
        self.assert_fact_repr(fact)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Factories for sqlalchemy models."""

import datetime

import factory
import faker
from nark.backends.sqlalchemy.objects import (
    AlchemyActivity,
    AlchemyCategory,
    AlchemyFact,
    AlchemyTag
)

from . import common


class AlchemyCategoryFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory class for generic ``AlchemyCategory`` instances."""

    # (lb): factoryboy starts step.sequence at 0,
    #   and we want positive PKs, so add one.
    pk = factory.Sequence(lambda n: n + 1)

    @factory.sequence
    def name(n):  # NOQA
        """Return a name that is guaranteed to be unique."""
        return '{name} - {key}'.format(name=faker.Faker().word(), key=n)

    deleted = False
    hidden = False

    class Meta:
        model = AlchemyCategory
        sqlalchemy_session = common.Session
        sqlalchemy_session_persistence = 'flush'


class AlchemyActivityFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory class for generic ``AlchemyActivity`` instances."""

    pk = factory.Sequence(lambda n: n + 1)
    name = factory.Faker('sentence')
    category = factory.SubFactory(AlchemyCategoryFactory)
    deleted = False
    hidden = False

    class Meta:
        model = AlchemyActivity
        sqlalchemy_session = common.Session
        sqlalchemy_session_persistence = 'flush'


class AlchemyTagFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory class for generic ``AlchemyTag`` instances."""

    pk = factory.Sequence(lambda n: n + 1)

    @factory.sequence
    def name(n):  # NOQA
        """Return a name that is guaranteed to be unique."""
        return '{name} - {key}'.format(name=faker.Faker().word(), key=n)

    deleted = False
    hidden = False

    class Meta:
        model = AlchemyTag
        sqlalchemy_session = common.Session
        sqlalchemy_session_persistence = 'flush'


class AlchemyFactFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory class for generic ``AlchemyFact`` instances."""

    pk = factory.Sequence(lambda n: n + 1)
    activity = factory.SubFactory(AlchemyActivityFactory)
    start = factory.Faker('date_time')
    end = factory.LazyAttribute(lambda o: o.start + datetime.timedelta(hours=3))
    description = factory.Faker('paragraph')

    class Meta:
        model = AlchemyFact
        sqlalchemy_session = common.Session
        sqlalchemy_session_persistence = 'flush'

    deleted = False
    split_from = None

    @factory.post_generation
    def tags(self, create, extracted, **kwargs):
        """Add new random tags after instance creation."""
        self.tags = list([AlchemyTagFactory() for i in range(4)])




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import logging

from nark.helpers import logging as logging_helpers


class TestSetupHandler(object):
    def test_get_formatter_basic(self, mocker):
        """Test formatter fetcher."""
        formatter = logging_helpers.formatter_basic()
        # (lb): Is this legit, or a little too _intimate?
        expected = '[%(levelname)s] %(asctime)s %(name)s %(funcName)s: %(message)s'
        assert formatter._fmt == expected

    def test_setup_handler_stream_handler(self, mocker):
        """Test logging setup."""
        stream_handler = logging.StreamHandler()
        formatter = logging_helpers.formatter_basic()
        logger = mocker.MagicMock()
        logging_helpers.setup_handler(stream_handler, formatter, logger)
        logger.addHandler.assert_called_with(stream_handler)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os

import lazy_import
from ....helpers.legacy_db import upgrade_legacy_db_hamster_applet
from ....managers.migrate import BaseMigrationsManager

# Profiling: Loading `migrate` takes ~ 0.090 seconds.
migrate_exceptions = lazy_import.lazy_module(
    'sqlalchemy_migrate_hotoffthehamster.exceptions'
)
migrate_versioning_cfgparse = lazy_import.lazy_module(
    'sqlalchemy_migrate_hotoffthehamster.versioning.cfgparse'
)
migrate_versioning_api = lazy_import.lazy_module(
    'sqlalchemy_migrate_hotoffthehamster.versioning.api'
)

__all__ = ('MigrationsManager', )


class MigrationsManager(BaseMigrationsManager):
    """"""

    @property
    def basepath(self):
        return self.migrations_path()

    @property
    def config(self):
        return self.migrations_config()

    # ***

    def control(self, version=None):
        """Mark a database as under version control."""
        current_ver = self.version()
        if current_ver is None:
            url = self.store.db_url
            try:
                migrate_versioning_api.version_control(
                    url, self.basepath, version=version, config=self.config,
                )
                return True
            except migrate_exceptions.DatabaseAlreadyControlledError:
                return False
        elif current_ver == 0:
            return False
        else:
            return None

    def downgrade(self):
        """Downgrade the database according to its migration version."""
        current_ver = self.version()
        if current_ver is None:
            return None
        latest_ver = migrate_versioning_api.version(
            self.basepath, config=self.config,
        )
        if not latest_ver:
            return None
        assert current_ver <= latest_ver
        if current_ver > 0:
            next_version = current_ver - 1
            url = self.store.db_url
            migrate_versioning_api.downgrade(
                url, self.basepath, version=next_version, config=self.config,
            )
            return True
        else:
            return False

    def upgrade(self):
        """Upgrade the database according to its migration version."""
        current_ver = self.version()
        if current_ver is None:
            return None
        latest_ver = migrate_versioning_api.version(
            self.basepath, config=self.config,
        )
        if not latest_ver:
            return None
        assert current_ver <= latest_ver
        if current_ver < latest_ver:
            next_version = current_ver + 1
            url = self.store.db_url
            migrate_versioning_api.upgrade(
                url, self.basepath, version=next_version, config=self.config,
            )
            return True
        else:
            return False

    def version(self):
        """Returns the current migration of the active database."""
        url = self.store.db_url
        try:
            return migrate_versioning_api.db_version(
                url, self.basepath, config=self.config,
            )
        except migrate_exceptions.DatabaseNotControlledError:
            return None

    def latest_version(self):
        """Returns the latest version defined by the application."""
        try:
            repo_latest = migrate_versioning_api.version(
                self.basepath, config=self.config,
            )
            return int(repo_latest.value)
        except migrate_exceptions.DatabaseNotControlledError:
            return None

    # ***

    def migrations_config(self):
        """Return sqlalchemy-migrate migrate.cfg a hot-n-ready config object."""
        # See nark/migrations/migrate.cfg.example for descriptions.
        config = migrate_versioning_cfgparse.Parser()
        config['db_settings'] = {}
        config['db_settings']['repository_id'] = 'nark-migrations'
        config['db_settings']['version_table'] = 'migrate_version'
        config['db_settings']['required_dbs'] = '[]'
        config['db_settings']['use_timestamp_numbering'] = 'False'
        return config

    def migrations_path(self):
        # (lb): This is a little awkward. But there's not
        # another convenient way to do this, is there?
        path = os.path.abspath(
            os.path.join(
                # Meh. We could also do dirname(nark.__file__) and use fewer ..'s.
                #    path = os.path.join(os.path.dirname(nark.__file__), 'migrations')
                # except that nark is not a package, it's this package!
                os.path.dirname(__file__),
                '../../../migrations',
            )
        )
        return path

    # ***

    def legacy_upgrade_from_hamster_applet(self, db_path):
        upgrade_legacy_db_hamster_applet(db_path)

    def legacy_upgrade_from_hamster_lib(self):
        # (lb): I'm not sure how much traction hamster-lib had.
        #  So I'm not sure this function is needed.
        #  And hamster-libbers out there need a hand?
        #  You'll need to rename 3 things is all, I'm sure:
        #    Table renamed: facttags → fact_tags
        #    Cols renamed: facts.start_time/end_time → facts.start/end
        #  Please open a ticket if you really want this feature!
        raise NotImplementedError




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from . import BaseManager
from ..items.category import Category


class BaseCategoryManager(BaseManager):
    """
    Base class defining the minimal API for a CategoryManager implementation.
    """

    def __init__(self, *args, **kwargs):
        super(BaseCategoryManager, self).__init__(*args, **kwargs)

    # ***

    def save(self, category):
        """
        Save a Category to the backend.

        Args:
            category (nark.Category): Category instance to be saved.

        Returns:
            nark.Category: Saved Category

        Raises:
            TypeError: If the ``category`` parameter is not a valid
                ``Category`` instance.
        """
        return super(BaseCategoryManager, self).save(category, Category, named=True)

    # ***

    def get_or_create(self, category):
        """
        Check if we already got a category with that name, if not create one.

        This is a convenience method as it seems sensible to rather implement
        this once in our controller than having every client implementation
        deal with it anew.

        It is worth noting that the lookup completely ignores any PK contained in the
        passed category. This makes this suitable to just create the desired Category
        and pass it along. One way or the other one will end up with a persisted
        db-backed version.

        Args:
            category (nark.Category or None): The categories.

        Returns:
            nark.Category or None: The retrieved or created category. Either way,
                the returned Category will contain all data from the backend, including
                its primary key.
        """

        self.store.logger.debug(_("'{}' has been received.'.".format(category)))
        if category:
            try:
                category = self.get_by_name(category)
            except KeyError:
                category = Category(category)
                category = self._add(category)
        else:
            # We want to allow passing ``category=None``, so we normalize here.
            category = None
        return category

    # ***

    def _add(self, category):
        """
        Add a ``Category`` to our backend.

        Args:
            category (nark.Category): ``Category`` to be added.

        Returns:
            nark.Category: Newly created ``Category`` instance.

        Raises:
            ValueError: When the category name was already present!
                It is supposed to be unique.
            ValueError: If category passed already got an PK.
                Indicating that update would be more appropriate.

        Note:
            * Legacy version stored the proper name as well as a ``lower(name)`` version
            in a dedicated field named ``search_name``.
        """
        raise NotImplementedError

    # ***

    def _update(self, category):
        """
        Update a ``Categories`` values in our backend.

        Args:
            category (nark.Category): Category to be updated.

        Returns:
            nark.Category: The updated Category.

        Raises:
            KeyError: If the ``Category`` can not be found by the backend.
            ValueError: If the ``Category().name`` is already being used by
                another ``Category`` instance.
            ValueError: If category passed does not have a PK.
        """
        raise NotImplementedError

    # ***

    def remove(self, category):
        """
        Remove a category.

        Any ``Activity`` referencing the passed category will be set to
        ``Activity().category=None``.

        Args:
            category (nark.Category): Category to be updated.

        Returns:
            None: If everything went ok.

        Raises:
            KeyError: If the ``Category`` can not be found by the backend.
            TypeError: If category passed is not an nark.Category instance.
            ValueError: If category passed does not have an pk.
        """
        raise NotImplementedError

    # ***

    def get(self, pk):
        """
        Get an ``Category`` by its primary key.

        Args:
            pk (int): Primary key of the ``Category`` to be fetched.

        Returns:
            nark.Category: ``Category`` with given primary key.

        Raises:
            KeyError: If no ``Category`` with this primary key can be found
                by the backend.
        """

        raise NotImplementedError

    # ***

    def get_by_name(self, name):
        """
        Look up a category by its name.

        Args:
            name (str): Unique name of the ``Category`` to we want to fetch.

        Returns:
            nark.Category: ``Category`` with given name.

        Raises:
            KeyError: If no ``Category`` with this name was found by the backend.
        """
        raise NotImplementedError

    # ***

    def get_all(
        self,
        include_usage=True,
        count_results=False,
        since=None,
        until=None,
        deleted=False,
        hidden=False,
        key=None,
        search_term='',
        activity=False,
        sort_col='',
        sort_order='',
        limit='',
        offset='',
    ):
        """
        Return a list of all categories.

        Returns:
            list: List of ``Categories``, ordered by ``lower(name)``.
        """
        raise NotImplementedError




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""``nark`` object store."""

from gettext import gettext as _

import os.path

# Profiling: load create_engine: ~ 0.100 secs.
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError
# Profiling: load sessionmaker: ~ 0.050 secs.
from sqlalchemy.orm import sessionmaker

from . import objects
from ...manager import BaseStore
from .managers.activity import ActivityManager
from .managers.category import CategoryManager
from .managers.fact import FactManager
from .managers.migrate import MigrationsManager
from .managers.tag import TagManager

__all__ = ('SQLAlchemyStore', )


class SQLAlchemyStore(BaseStore):
    """
    SQLAlchemy based backend.

    Unfortunately despite using SQLAlchemy some database specific settings can
    not be avoided (autoincrement, indexes etc).

    Some of those issues will not be relevant in later versions as we
    may get rid of Category and Activity ids entirely, just using their
    natural/composite keys as primary keys.

    However, for now we just support sqlite until the basic framework is up
    and running. It should take only minor but delayable effort to broaden the
    applicability to postgres, mysql and the likes.

    The main takeaway right now is, that their is no actual guarantee that
    in a distributed environment no race condition occur and we may end up
    with duplicate Category/Activity entries. No backend code will be able to
    prevent this by virtue of this being a DB issue.

    Furthermore, we will try hard to avoid placing more than one fact in a
    given time window. However, there can be no guarantee that in a distributed
    environment this will always work out. As a consequence, we make sure that
    all our single object data retrieval methods return only one item or throw
    an error alerting us about the inconsistency.
    """

    def __init__(self, config):
        """
        """
        super(SQLAlchemyStore, self).__init__(config)
        self.create_item_managers()

    def standup(self, session=None):
        """
        Set up the store.

        Args:
            config (dict): Dictionary of config key/value pairs.

            session (SQLALcheny Session object, optional): Provide a dedicated session
                to be used. Defaults to ``None``.

        Note:
            The ``session`` argument is mainly useful for tests.
        """
        engine = self.create_storage_engine()
        created_fresh = self.create_storage_tables(engine)
        self.initiate_storage_session(session, engine)
        if created_fresh:
            self.control_and_version_store()
        return created_fresh

    def cleanup(self):
        pass

    @property
    def db_url(self):
        """
        Create a ``database_url`` from ``config`` suitable to be consumed
        by ``create_engine``

        Our config may include:
            * ''db.engine``; Engine to be used.
            * ``db.host``; Host to connect to.
            * ``db.port``; Port to connect to.
            * ``db.path``; Used if ``engine='sqlite'``.
            * ``db.user``; Database user to be used for connection.
            * ``db.password``; Database user passwort to authenticate user.

        If ``db.engine='sqlite'`` you need to provide ``db.path`` as well.
        For any other engine ``db.host`` and ``db.name`` are mandatory.

        Note:
            * `SQLAlchemy docs
              <https://docs.sqlalchemy.org/en/latest/core/engines.html>`__

        Returns:
            str: ``database_url`` suitable to be consumed by ``create_engine``.

        Raises:
            ValueError: If a required config key/value pair is not present for
                the choosen ``db.engine``.
        """
        # [FIXME]
        # Contemplate if there are security implications that warrant sanitizing
        # config values.
        # (lb): This is an old hamster-lib comment. The user will be the ones
        # deliberately setting their config, so if their was a security concern
        # (SQL injection?), the user will have known what they've done.
        # - Can probably delete both these comments now.

        # Because ConfigDecorator, we could access variables with
        # dot-notation or by subscripting, e.g., these both work:
        #   engine = self.config['db.engine']
        #   engine = self.config['db']['engine']
        # The code generally uses dot-notation because that's how
        # the user sees it in `dob config dump` output; it's more
        # concise; and it's easier to bang out in a search command.
        engine = self.config['db.engine']
        host = self.config['db.host']
        name = self.config['db.name']
        path = self.config['db.path']
        port = self.config['db.port']
        user = self.config['db.user']
        password = self.config['db.password']

        if not engine:
            message = _("No engine found in config!")
            self.logger.error(message)
            raise ValueError(message)

        # URL composition is slightly different for sqlite
        if engine == 'sqlite':
            if not path:
                # We could have allowed for blank paths, which would make
                # SQLAlchemy default to ``:memory:``. But explicit is better
                # than implicit. You can still create an in memory db by passing
                # ``db_path=':memory:'`` deliberately.
                message = _("No 'db.path' found in config! Sqlite requires one.")
                self.logger.error(message)
                raise ValueError(message)
            if path != ':memory:':
                # Make sure we always use an absolute path.
                path = os.path.abspath(path)
            database_url = '{engine}:///{path}'.format(engine=engine, path=path)
        else:
            if not host:
                message = _(
                    "No 'db.host' found in config!"
                    " Engines other than sqlite require one."
                )
                self.logger.error(message)
                raise ValueError(message)
            if not name:
                message = _(
                    "No 'db.name' found in config!"
                    " Engines other than sqlite require one."
                )
                self.logger.error(message)
                raise ValueError(message)
            if not user:
                message = _(
                    "No 'db.user' found in config!"
                    " Engines other than sqlite require one."
                )
                self.logger.error(message)
                raise ValueError(message)
            if not password:
                message = _(
                    "No 'db.password' found in config!"
                    " Engines other than sqlite require one."
                )
                self.logger.error(message)
                raise ValueError(message)
            if port:
                port = ':{}'.format(port)
            database_url = (
                '{engine}://{user}:{password}@{host}{port}/{name}'
                .format(
                    engine=engine,
                    user=user,
                    password=password,
                    host=host,
                    port=port,
                    name=name,
                )
            )
        self.logger.debug(_('database_url: {}'.format(database_url)))
        return database_url

    def create_storage_engine(self):
        # [TODO]
        # It takes more deliberation to decide how to handle engine creation
        # if we receive a session. Should be require the session to bring
        # its own engine?
        engine = create_engine(self.db_url)
        self.logger.debug(_('Engine created.'))
        # NOTE: (lb): I succeeded at setting the ORM (Sqlite3) logger level,
        # but it didn't log anything (I was hoping to see all statements).
        #
        #  import logging
        #  engine.logger.setLevel(logging.DEBUG)
        return engine

    def create_storage_tables(self, engine):
        objects.metadata.bind = engine
        created_fresh = False
        try:
            # Create the database store at db_path.
            objects.metadata.create_all(engine, checkfirst=False)
            created_fresh = True
            self.logger.debug(_("Database tables created."))
        except OperationalError:
            # E.g., '(sqlite3.OperationalError) table categories already exists'.
            self.logger.debug(_("Database tables already exist."))
        return created_fresh

    def control_and_version_store(self):
        # (lb): I'm not sure how else to do this, or if this is "the way":
        #         Put the new database under migration control.
        latest_vers = self.migrations.latest_version()
        assert latest_vers is not None and latest_vers > 0
        self.migrations.control(version=latest_vers)

    def initiate_storage_session(self, session, engine):
        if not session:
            Session = sessionmaker(bind=engine)  # NOQA
            self.logger.debug(_("Bound engine to session-object."))
            self.session = Session()
            self.logger.debug(_("Instantiated session."))
        else:
            self.session = session

    def create_item_managers(self):
        self.migrations = MigrationsManager(self)
        self.categories = CategoryManager(self)
        self.activities = ActivityManager(self)
        self.tags = TagManager(self)
        localize = self.config['time.tz_aware']
        self.facts = FactManager(self, localize=localize)
        self.fact_cls = None




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Base class for Nark item instances."""

__all__ = ('BaseItem', )


class BaseItem(object):
    """Base class for all items."""

    def __init__(self, pk, name):
        self.pk = pk
        self.name = name

    def __repr__(self):
        parts = []
        for key in sorted(self.__dict__.keys()):
            parts.append(
                "{key}={val}".format(key=key, val=repr(getattr(self, key)))
            )
        repred = "{cls}({parts})".format(
            cls=self.__class__.__name__, parts=', '.join(parts),
        )
        return repred

    @property
    def unstored(self):
        return (not self.pk) or (self.pk < 0)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import csv
import datetime
import os.path
import xml

import pytest
from icalendar import Calendar
from nark import reports


# Fixtures
@pytest.fixture
def path(tmpdir):
    path = tmpdir.mkdir('reports').join('report.csv').strpath
    return path


@pytest.fixture
def report_writer(path):
    return reports.ReportWriter(path)


@pytest.fixture
def tsv_writer(path):
    return reports.TSVWriter(path)


@pytest.fixture
def ical_writer(path):
    return reports.ICALWriter(path)


@pytest.fixture
def xml_writer(path):
    return reports.XMLWriter(path)


# Tests
class TestReportWriter(object):
    @pytest.mark.parametrize('datetime_format', [None, '%Y-%m-%d'])
    def test_init_stores_datetime_format(self, path, datetime_format):
        """Make sure that Writer initialization stores the ``datetime_format``."""
        writer = reports.ReportWriter(path, datetime_format)
        assert writer.datetime_format == datetime_format

    def test_init_file_opened(self, path):
        """Make sure a file like object is beeing opened."""
        writer = reports.ReportWriter(path)
        assert os.path.isfile(path)
        assert writer.file.closed is False

    def test__fact_to_tuple(self, report_writer, fact):
        with pytest.raises(NotImplementedError):
            report_writer._fact_to_tuple(fact)

    def test_write_report_write_lines(self, mocker, report_writer, list_of_facts):
        """Make sure that each ``Fact`` instances triggers a new line."""
        number_of_facts = 10
        facts = list_of_facts(number_of_facts)
        report_writer._write_fact = mocker.MagicMock(return_value=None)
        report_writer._fact_to_tuple = mocker.MagicMock(return_value=None)
        report_writer.write_report(facts)
        assert report_writer._write_fact.call_count == number_of_facts

    def test_write_report_file_closed(self, report_writer, list_of_facts):
        """Make sure our output file is closed at the end."""
        facts = list_of_facts(10)
        with pytest.raises(NotImplementedError):
            report_writer.write_report(facts)
        assert report_writer.file.closed is False

    def test__close(self, report_writer, path):
        """Ensure that the the output gets closed."""
        report_writer._close()
        assert report_writer.file.closed


class TestTSVWriter(object):
    def test_init_csv_writer(self, tsv_writer):
        """Make sure that initialition provides us with a ``csv.writer`` instance."""
        assert tsv_writer.csv_writer
        assert tsv_writer.csv_writer.dialect == csv.get_dialect('excel-tab')

    def test_init_heading(self, path, tsv_writer):
        """Make sure that initialition writes header as expected."""

        expectations = (
            'start time',
            'end time',
            'duration minutes',
            'activity',
            'category',
            'description',
            'deleted',
        )

        tsv_writer._close()
        with open(path, 'r') as fobj:
            reader = csv.reader(fobj, dialect='excel-tab')
            header = next(reader)
        for field, expectation in zip(header, expectations):
            if isinstance(field, str):
                assert field == expectation
            else:
                assert field.decode('utf-8') == expectation

    def test__fact_to_tuple_no_category(self, tsv_writer, fact):
        """Make sure that ``None`` category values translate to ``empty strings``."""
        fact.activity.category = None
        result = tsv_writer._fact_to_tuple(fact)
        assert result.category == ''

    def test__fact_to_tuple_with_category(self, tsv_writer, fact):
        """Make sure that category references translate to their names."""
        result = tsv_writer._fact_to_tuple(fact)
        assert result.category == fact.category.name

    def test__write_fact(self, path, fact, tsv_writer):
        """Make sure the writen fact is what we expect."""
        fact_tuple = tsv_writer._fact_to_tuple(fact)
        tsv_writer._write_fact(fact_tuple)
        tsv_writer._close()
        with open(path, 'r') as fobj:
            reader = csv.reader(fobj, dialect='excel-tab')
            next(reader)
            line = next(reader)
            for field, expectation in zip(line, fact_tuple):
                if isinstance(field, str):
                    assert field == expectation
                else:
                    assert field.decode('utf-8') == expectation


class TestICALWriter(object):
    """Make sure the iCal writer works as expected."""
    def test_init(self, ical_writer):
        """Make sure that init creates a new calendar instance to add events to."""
        assert ical_writer.calendar

    def test__fact_to_tuple(self, ical_writer, fact):
        """Make sure that our general expection about conversions are matched."""
        result = ical_writer._fact_to_tuple(fact)
        assert result.start == fact.start
        assert result.end == fact.end
        assert result.activity == fact.activity.name
        assert result.duration is None
        assert result.category == fact.category.name
        assert result.description == fact.description

    def test__fact_to_tuple_no_category(self, ical_writer, fact):
        """Make sure that ``None`` category values translate to ``empty strings``."""
        fact.activity.category = None
        result = ical_writer._fact_to_tuple(fact)
        assert result.category == ''

    def test__fact_to_tuple_with_category(self, ical_writer, fact):
        """Make sure that category references translate to their names."""
        result = ical_writer._fact_to_tuple(fact)
        assert result.category == fact.category.name

    def test_write_fact(self, ical_writer, fact, mocker):
        """Make sure that the fact attached to the calendar matches our expectations."""
        fact_tuple = ical_writer._fact_to_tuple(fact)
        ical_writer.calendar.add_component = mocker.MagicMock()
        ical_writer._write_fact(fact_tuple)
        result = ical_writer.calendar.add_component.call_args[0][0]
        assert result.get('dtstart').dt == fact_tuple.start
        assert result.get('dtend').dt == fact_tuple.end + datetime.timedelta(seconds=1)
        assert result.get('summary') == fact_tuple.activity
        # Make lists of [vText] and [str], else comparison fails.
        #  NO: assert result.get('categories') == fact_tuple.category
        assert list(result.get('categories').cats) == list(fact_tuple.category)
        assert result.get('description') == fact_tuple.description

    def test__close(self, ical_writer, fact, path):
        """Make sure the calendar is actually written do disk before file is closed."""
        ical_writer.write_report((fact,))
        with open(path, 'rb') as fobj:
            result = Calendar.from_ical(fobj.read())
            assert result.walk()


class TestXMLWriter(object):
    """Make sure the XML writer works as expected."""

    def test_init_(self, xml_writer):
        """Make sure a XML main document and a facts list child element is set up."""
        assert xml_writer.document
        assert xml_writer.fact_list

    def test_fact_to_tuple(self, xml_writer, fact):
        """Make sure type conversion and normalization matches our expectations."""
        result = xml_writer._fact_to_tuple(fact)
        assert result.start == fact.start.strftime(xml_writer.datetime_format)
        assert result.end == fact.end.strftime(xml_writer.datetime_format)
        assert result.activity == fact.activity.name
        assert result.duration == fact.format_delta(style='%M')
        assert result.category == fact.category.name
        assert result.description == fact.description

    def test__fact_to_tuple_no_category(self, xml_writer, fact):
        """Make sure that ``None`` category values translate to ``empty strings``."""
        fact.activity.category = None
        result = xml_writer._fact_to_tuple(fact)
        assert result.category == ''

    def test_write_fact(self, xml_writer, fact, mocker):
        """Make sure that the attributes attached to the fact matche our expectations."""
        fact_tuple = xml_writer._fact_to_tuple(fact)
        xml_writer.fact_list.appendChild = mocker.MagicMock()
        xml_writer._write_fact(fact_tuple)
        result = xml_writer.fact_list.appendChild.call_args[0][0]
        assert result.getAttribute('start') == fact_tuple.start
        assert result.getAttribute('end') == fact_tuple.end
        assert result.getAttribute('duration') == fact_tuple.duration
        assert result.getAttribute('activity') == fact_tuple.activity
        assert result.getAttribute('category') == fact_tuple.category
        assert result.getAttribute('description') == fact_tuple.description

    def test__close(self, xml_writer, fact, path):
        """Make sure the calendar is actually written do disk before file is closed."""
        xml_writer.write_report((fact,))
        with open(path, 'rb') as fobj:
            result = xml.dom.minidom.parse(fobj)
            assert result.toxml()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module defines logging related function."""

from gettext import gettext as _

import logging

from ansi_escape_room import attr, fg


def formatter_basic(color=False):
    if not color:
        return formatter_basic_plain()
    return formatter_basic_color()


def formatter_basic_plain():
    formatter = logging.Formatter(
        '[%(levelname)s] '
        '%(asctime)s '
        '%(name)s '
        '%(funcName)s: '
        '%(message)s'
    )
    return formatter


def formatter_basic_color():
    formatter = logging.Formatter(
        '{grey_54}[{underlined}{magenta}%(levelname)s{reset}{grey_54}]{reset} '
        '{yellow}%(asctime)s{reset} '
        '{light_blue}%(name)s '
        '%(funcName)s{reset}: '
        '{bold}{green}%(message)s{reset}'.format(
            # MAGIC: Use RGB not 'grey_54' because tmux 0-255 color issue.
            #   grey_54=fg('grey_54'),
            grey_54=fg('#8a8a8a'),
            underlined=attr('underlined'),
            magenta=fg('magenta'),
            reset=attr('reset'),
            yellow=fg('yellow'),
            light_blue=fg('light_blue'),
            bold=attr('bold'),
            green=fg('green'),
        )
    )
    return formatter


def resolve_log_level(level):
    error = False
    try:
        try:
            log_level = int(level)
        except ValueError:
            log_level = logging.getLevelName(level)
    except KeyError:
        error = True
        log_level = logging.WARNING
    return log_level, error


def set_logger_level(logger_name, logger_log_level):
    logger = logging.getLogger(logger_name)
    logger.addHandler(logging.NullHandler())

    # (lb): BIZARRE: On a 14.04 machine, parent.handlers has StreamHandler
    #   in it, so it prints to console. This does not happen on a 16.04
    #   machine I also use. And I cannot determine the reason (both
    #   machines use a virtualenv configured exactly the same way, and
    #   the Python version is merely off by one PATCH).
    logger.parent.handlers = []
    logger.parent.addHandler(logging.NullHandler())

    log_level, warn_name = resolve_log_level(logger_log_level)

    try:
        logger.setLevel(int(log_level))
    except ValueError:
        warn_name = True
        logger.setLevel(logging.WARNING)

    if warn_name:
        logger.warning(
            _('Unknown log_level specified for ‘{}’: {}')
            .format(logger_name, logger_log_level)
        )

    return logger


def setup_handler(handler, formatter, *loggers):
    handler.setFormatter(formatter)
    for logger in loggers:
        logger.addHandler(handler)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

__all__ = (
    'Singleton',
)


class Singleton(type):
    """A Singleton metaclass.

    For a healthy discussion on ways to implement Singleton in Python,
    and whether or not they're a good tool to use, read the long-standing
    and still-rolling *Creating a singleton in Python* article:

        https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures needed to test helper submodule."""

import os

import pytest

from nark.helpers.app_dirs import NarkAppDirs


@pytest.fixture
def appdirs(mocker, tmpdir):
    """Provide mocked version specific user dirs using a tmpdir."""
    def ensure_directory_exists(directory):
        if not os.path.lexists(directory):
            os.makedirs(directory)
        return directory

    NarkAppDirs.user_config_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('config').strpath, 'nark/'))
    NarkAppDirs.user_data_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('data').strpath, 'nark/'))
    NarkAppDirs.user_cache_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('cache').strpath, 'nark/'))
    NarkAppDirs.user_log_dir = ensure_directory_exists(os.path.join(
        tmpdir.mkdir('log').strpath, 'nark/'))
    return NarkAppDirs


@pytest.fixture(params=[
    ('foobar', {
        'timeinfo': None,
        'activity': 'foobar',
        'category': None,
        'description': None,
    }),
    ('11:00 12:00 foo@bar', {
        'timeinfo': '11:00',
        'activity': '12:00 foo',
        'category': 'bar',
        'description': None,
    }),
    ('rumpelratz foo@bar', {
        'timeinfo': None,
        'start': None,
        'end': None,
        'activity': 'rumpelratz foo',
        'category': 'bar',
        'description': None,
    }),
    ('foo@bar', {
        'timeinfo': '',
        'activity': 'foo',
        'category': 'bar',
        'description': None,
    }),
    ('foo@bar, palimpalum', {
        'timeinfo': None,
        'activity': 'foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
    ('12:00 foo@bar, palimpalum', {
        'timeinfo': '12:00',
        'activity': 'foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
    ('12:00 - 14:14 foo@bar, palimpalum', {
        'timeinfo': '12:00 to 14:14',
        'activity': 'foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
    # Missing whitespace around ``-`` will prevent timeinfo from being parsed.
    ('12:00-14:14 foo@bar, palimpalum', {
        'timeinfo': '',
        'activity': '12:00-14:14 foo',
        'category': 'bar',
        'description': 'palimpalum',
    }),
])
# FIXME/2020-01-31: raw_fact_parametrized is not used.
def raw_fact_parametrized(request):
    """Provide a variety of raw facts as well as a dict of its proper components."""
    return request.param




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os

import pytest

from nark.config import decorate_config
from nark.helpers.app_dirs import NarkAppDirs


class TestNarkAppDirs(object):
    """Make sure that our custom AppDirs works as intended."""

    def test_user_data_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.user_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.user_data_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_data_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_data_dir) is create

    def test_site_data_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.site_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.site_data_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_site_data_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.site_data_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.site_data_dir) is create

    def test_user_config_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch(
            'nark.helpers.app_dirs.appdirs.user_config_dir',
            return_value=path,
        )
        appdir = NarkAppDirs('nark')
        assert appdir.user_config_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_config_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_config_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_config_dir) is create

    def test_site_config_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.site_config_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.site_config_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_site_config_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.site_config_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.site_config_dir) is create

    def test_user_cache_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.user_cache_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.user_cache_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_cache_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_cache_dir',
                     return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_cache_dir) is create

    def test_user_log_dir_returns_directoy(self, tmpdir, mocker):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        mocker.patch('nark.helpers.app_dirs.appdirs.user_log_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        assert appdir.user_log_dir == path

    @pytest.mark.parametrize('create', [True, False])
    def test_user_log_dir_creates_file(self, tmpdir, mocker, create, faker):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        mocker.patch('nark.helpers.app_dirs.appdirs.user_log_dir', return_value=path)
        appdir = NarkAppDirs('nark')
        appdir.create = create
        assert os.path.exists(appdir.user_log_dir) is create


class TestConfigObjToBackendConfig(object):
    """Make sure that conversion works expected."""

    def test_regular_usecase(self, configobj_instance):
        """Make sure basic mechanics work and int/time types are created."""
        configobj, expectation = configobj_instance
        result = decorate_config(configobj).as_dict(unmutated=True)
        assert result == expectation




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from . import BaseManager
from ..items.activity import Activity


class BaseActivityManager(BaseManager):
    """
    Base class defining the minimal API for a ActivityManager implementation.
    """
    def __init__(self, *args, **kwargs):
        super(BaseActivityManager, self).__init__(*args, **kwargs)

    # ***

    def save(self, activity):
        """
        Save an ``Activity`` to the backend.

        Args:
            activity (nark.Activity): ``Activity`` to be saved.

        Returns:
            nark.Activity: The saved ``Activity``.

        Raises:
            TypeError: If the ``activity`` parameter is not a valid
                ``Activity`` instance.
        """
        return super(BaseActivityManager, self).save(activity, Activity, named=True)

    # ***

    def get_or_create(self, activity):
        """
        Convenience method to either get an activity matching the specs
            or create a new one.

        Args:
            activity (nark.Activity): The activity we want.

        Returns:
            nark.Activity: The retrieved or created activity
        """
        self.store.logger.debug(_("'{}' has been received.".format(activity)))
        try:
            activity = self.get_by_composite(activity.name, activity.category)
        except KeyError:
            activity = self.save(
                Activity(
                    activity.name,
                    category=activity.category,
                    deleted=bool(activity.deleted),
                )
            )
        return activity

    # ***

    def _add(self, activity):
        """
        Add a new ``Activity`` instance to the database.

        Args:
            activity (nark.Activity): The ``Activity`` to be added.

        Returns:
            nark.Activity: The newly created ``Activity``.

        Raises:
            ValueError: If the passed activity has a PK.
            ValueError: If the category/activity.name combination to be added is
                already present in the db.

        Note:
            According to ``storage.db.Storage.__add_activity``: when adding
            a new activity with a new category, this category does not get
            created but instead this activity.category=None. This makes sense
            as categories passed are just ids, we however can pass full
            category objects. At the same time, this approach allows to add
            arbitrary category.id as activity.category without checking their
            existence. This may lead to db anomalies.
        """
        raise NotImplementedError

    # ***

    def _update(self, activity):
        """
        Update values for a given activity.

        Which activity to refer to is determined by the passed PK new values
        are taken from passed activity as well.

        Args:
            activity (nark.Activity): Activity to be updated.

        Returns:
            nark.Activity: Updated activity.
        Raises:
            ValueError: If the new name/category.name combination is already taken.
            ValueError: If the the passed activity does not have a PK assigned.
            KeyError: If the the passed activity.pk can not be found.

        Note:
            Seems to modify ``index``.
        """

        raise NotImplementedError

    # ***

    def remove(self, activity):
        """
        Remove an ``Activity`` from the database.import

        If the activity to be removed is associated with any ``Fact``-instances,
        we set ``activity.deleted=True`` instead of deleting it properly.
        If it is not, we delete it from the backend.

        Args:
            activity (nark.Activity): The activity to be removed.

        Returns:
            bool: True

        Raises:
            KeyError: If the given ``Activity`` can not be found in the database.

        Note:
            Should removing the last activity of a category also trigger category
            removal?
        """

        raise NotImplementedError

    # ***

    def get(self, pk):
        """
        Return an activity based on its primary key.

        Args:
            pk (int): Primary key of the activity

        Returns:
            nark.Activity: Activity matching primary key.

        Raises:
            KeyError: If the primary key can not be found in the database.
        """
        raise NotImplementedError

    # ***

    def get_by_composite(self, name, category):
        """
        Lookup for unique 'name/category.name'-composite key.

        This method utilizes that to return the corresponding entry or None.

        Args:
            name (str): Name of the ``Activities`` in question.

            category (nark.Category or None): ``Category`` of the activities.
                May be None.

        Returns:
            nark.Activity: The corresponding activity

        Raises:
            KeyError: If the composite key can not be found.
        """
        # [FIXME]
        # Handle resurrection. See legacy
        # ``nark.storage.db.__get_activity_by_name``

        raise NotImplementedError

    # ***

    def get_all(
        self,
        include_usage=True,
        count_results=False,
        since=None,
        until=None,
        deleted=False,
        hidden=False,
        key=None,
        search_term='',
        category=False,
        activity=False,
        sort_col='',
        sort_order='',
        limit='',
        offset='',
    ):
        """
        Return all matching activities.

        Args:
            category (nark.Category, optional): Limit activities to this category.
                Defaults to ``False``. If ``category=None`` only activities without a
                category will be considered.
            search_term (str, optional): Limit activities to those matching this string
                a substring in their name. Defaults to ``empty string``.

        Returns:
            list: List of ``nark.Activity`` instances matching constrains.
                This list is ordered by ``Activity.name``.

        Note:
            * This method combines legacy ``storage.db.__get_activities`` and
                ``storage.db.____get_category_activities``.
            * Can search terms be prefixed with 'not'?
            * Original implementation in ``nark.storage.db.__get_activities`` returns
                activity names converted to lowercase!
            * Does exclude activities with ``deleted=True``.
        """
        # [FIXME]
        # ``__get_category_activity`` order by lower(activity.name),
        # ``__get_activities```orders by most recent start date *and*
        # lower(activity.name).
        raise NotImplementedError

    # ***

    def get_all_by_usage(
        self,
        deleted=False,
        hidden=False,
        search_term='',
        category=False,
        activity=False,
        sort_col='',
        sort_order='',
        limit='',
        offset='',
    ):
        """
        Similar to get_all(), but include count of Facts that reference each Activity.
        """
        raise NotImplementedError




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark

"""
Packaging instruction for setup tools.

Refs:

  https://setuptools.readthedocs.io/

  https://packaging.python.org/en/latest/distributing.html

  https://github.com/pypa/sampleproject
"""

from setuptools import find_packages, setup

# *** Package requirements.

requirements = [
    # "Very simple Python library for color and formatting in terminal."
    # Forked by HOTH (for italic "support"):
    #  https://github.com/hotoffthehamster/ansi-escape-room
    # Forked from dslackw:
    #  https://gitlab.com/dslackw/colored
    # Used in nark to format log messages, but also installed by dob(-bright),
    #  so not really adding any overhead.
    'ansi-escape-room >= 1.4.2, < 2',
    # Platform-specific directory magic.
    #  https://github.com/ActiveState/appdirs
    'appdirs >= 1.4.3, < 2',
    # Pythonic config @decorator.
    #  https://github.com/hotoffthehamster/config-decorator
    'config_decorator >= 2.0.0, < 3',
    # Better INI/conf parser (preserves order, comments) than ConfigParser.
    #  https://github.com/DiffSK/configobj
    #  https://configobj.readthedocs.io/en/latest/
    'configobj >= 5.0.6, < 6',
    # https://github.com/scrapinghub/dateparser
    'dateparser >= 0.7.4, < 1',
    # Elapsed timedelta formatter, e.g., "1.25 days".
    # - Imports as `pedantic_timedelta`.
    #  https://github.com/hotoffthehamster/human-friendly_pedantic-timedelta
    'human-friendly_pedantic-timedelta >= 2.0.0, < 3',
    # https://github.com/collective/icalendar
    'icalendar >= 4.0.5, < 5',
    # https://bitbucket.org/micktwomey/pyiso8601
    'iso8601 >= 0.1.12, < 1',
    # https://github.com/mnmelo/lazy_import
    'lazy_import >= 0.2.2, < 1',
    # Daylight saving time-aware timezone library.
    #  https://pythonhosted.org/pytz/
    'pytz >= 2019.3',
    # For testing with dateparser,
    #   https://bitbucket.org/mrabarnett/mrab-regex
    #   https://pypi.org/project/regex/
    'regex >= 2020.2.20',
    # FIXME/2019-11-02: (lb): Should (probably) upgrade to SQLalchemy 1.3.
    #  https://docs.sqlalchemy.org/en/13/changelog/migration_13.html
    # https://www.sqlalchemy.org/
    'sqlalchemy >= 1.2.19, < 1.3',
    # Database gooser/versioner.
    #  https://pypi.org/project/sqlalchemy-migrate/
    #  https://sqlalchemy-migrate.readthedocs.io/en/latest/
    # 2019-02-21: (lb): Forked again! Package alt. that accepts static config.
    'sqlalchemy-migrate-hotoffthehamster >= 0.13.0, < 1',
    # https://github.com/regebro/tzlocal
    'tzlocal >= 2.0.0, < 3',
]

# *** Minimal setup() function -- Prefer using config where possible.

# (lb): Most settings are in setup.cfg, except identifying packages.
# (We could find-packages from within setup.cfg, but it's convoluted.)

setup(
    # Run-time dependencies installed on `pip install`. To learn more
    # about "install_requires" vs pip's requirements files, see:
    #   https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # Specify which package(s) to install.
    # - Without any rules, find_packages returns, e.g.,
    #     ['nark', 'tests', 'tests.nark']
    # - With the 'exclude*' rule, this call is essentially:
    #     packages=['nark']
    packages=find_packages(exclude=['tests*']),

    # Tell setuptools to determine the version
    # from the latest SCM (git) version tag.
    #
    # Note that if the latest commit is not tagged with a version,
    # or if your working tree or index is dirty, then the version
    # from git will be appended with the commit hash that has the
    # version tag, as well as some sort of 'distance' identifier.
    # E.g., if a project has a '3.0.0a21' version tag but it's not
    # on HEAD, or if the tree or index is dirty, the version might
    # be:
    #   $ python setup.py --version
    #   3.0.0a22.dev3+g6f93d8c.d20190221
    # But if you clean up your working directory and move the tag
    # to the latest commit, you'll get the plain version, e.g.,
    #   $ python setup.py --version
    #   3.0.0a31
    # Ref:
    #   https://github.com/pypa/setuptools_scm
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from collections import namedtuple

from .item_base import BaseItem

CategoryTuple = namedtuple(
    'CategoryTuple', ('pk', 'name', 'deleted', 'hidden'),
)


class Category(BaseItem):
    """Storage agnostic class for categories."""

    def __init__(self, name, pk=None, deleted=False, hidden=False):
        """
        Initialize this instance.

        Args:
            name (str): The name of the category. May contain whitespace!
            pk: The unique primary key used by the backend.
        """

        super(Category, self).__init__(pk, name)
        self.deleted = bool(deleted)
        self.hidden = bool(hidden)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            raise ValueError(_('Category name must not be None.'))
        self._name = str(name)

    def as_tuple(self, include_pk=True):
        """
        Provide a tuple representation of this categories relevant 'fields'.

        Args:
            include_pk (bool): Whether to include the instances pk or not. Note that if
            ``False`` ``tuple.pk = False``!

        Returns:
            CategoryTuple: Representing this categories values.
        """
        pk = self.pk
        if not include_pk:
            pk = False
        cat_tup = CategoryTuple(
            pk=pk,
            name=self.name,
            deleted=bool(self.deleted),
            hidden=bool(self.hidden),
        )
        return cat_tup

    def equal_fields(self, other):
        """
        Compare this instances fields with another category.
        This excludes comparing the PK.

        Args:
            other (Category): Category to compare this instance with.

        Returns:
            bool: ``True`` if all fields but ``pk`` are equal, ``False`` if not.

        Note:
            This is particularly useful if you want to compare a new
            ``Category`` instance with a freshly created backend instance. As
            the latter will probably have a primary key assigned now and so
            ``__eq__`` would fail.
        """
        if other:
            other = other.as_tuple(include_pk=False)
        else:
            other = None

        return self.as_tuple(include_pk=False) == other

    def __eq__(self, other):
        if other is not None and not isinstance(other, CategoryTuple):
            other = other.as_tuple()
        return self.as_tuple() == other

    def __hash__(self):
        """Naive hashing method."""
        return hash(self.as_tuple())

    def __str__(self):
        return '{name}'.format(name=self.name)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import datetime

from . import BaseManager
from ..helpers import fact_time
from ..items.fact import Fact


class BaseFactManager(BaseManager):
    """Base class defining the minimal API for a FactManager implementation."""
    def __init__(self, *args, localize=False, **kwargs):
        super(BaseFactManager, self).__init__(*args, **kwargs)
        # (lb): Setting a class variable makes me feel somewhat dirty. Somewhat.
        Fact.localize(localize)

    # ***

    def save(self, fact, **kwargs):
        """
        Save a Fact to our selected backend.

        Unlike the private ``_add`` and ``_update`` methods, ``save``
        requires that the config given ``fact_min_delta`` is enforced.

        Args:
            fact (nark.Fact): Fact to be saved. Needs to be complete otherwise
            this will fail.

        Returns:
            nark.Fact: Saved Fact.

        Raises:
            ValueError: If ``fact.delta`` is smaller than
              ``self.config['time.fact_min_delta']``
        """
        def _save():
            enforce_fact_min_delta()
            return super(BaseFactManager, self).save(
                fact, cls=Fact, named=False, **kwargs
            )

        def enforce_fact_min_delta():
            # BROKEN/DONT_CARE: (lb): The Facts Carousel does not check the
            # min delta, meaning you could violate fact_min_delta and end up
            # raising from herein. Oh, well, I don't delta, so I don't care.
            if not fact.end:
                # The ongoing fact.
                return

            fact_min_delta = int(self.config['time.fact_min_delta'])
            if not fact_min_delta:
                # User has not enabled min-delta behavior.
                return

            min_delta = datetime.timedelta(seconds=fact_min_delta)
            if fact.delta() >= min_delta:
                # Fact is at least as long as user's min-delta.
                return

            message = _(
                "The Fact duration is shorter than the mandatory value of "
                "{} seconds specified in your config.".format(fact_min_delta)
            )
            self.store.logger.error(message)
            raise ValueError(message)

        return _save()

    # ***

    def _add(self, fact):
        """
        Add a new ``Fact`` to the backend.

        Args:
            fact (nark.Fact): Fact to be added.

        Returns:
            nark.Fact: Added ``Fact``.

        Raises:
            ValueError: If passed fact has a PK. New facts should not have one.
            ValueError: If timewindow is already occupied.
        """
        raise NotImplementedError

    # ***

    def _update(self, fact):
        """
        Update and existing fact with new values.

        Args:
            fact (nark.fact): Fact instance holding updated values.

        Returns:
            nark.fact: Updated Fact

        Raises:
            KeyError: if a Fact with the relevant PK could not be found.
            ValueError: If the the passed activity does not have a PK assigned.
            ValueError: If the timewindow is already occupied.
        """
        raise NotImplementedError

    # ***

    def remove(self, fact, purge=False):
        """
        Remove a given ``Fact`` from the backend.

        Args:
            fact (nark.Fact): ``Fact`` instance to be removed.

        Returns:
            bool: Success status

        Raises:
            ValueError: If fact passed does not have an pk.
            KeyError: If the ``Fact`` specified could not be found in the backend.
        """
        raise NotImplementedError

    # ***

    def get(self, pk, deleted=None):
        """
        Return a Fact by its primary key.

        Args:
            pk (int): Primary key of the ``Fact to be retrieved``.

            deleted (boolean, optional): False to restrict to non-deleted
                Facts; True to find only those marked deleted; None to find
                all.

        Returns:
            nark.Fact: The ``Fact`` corresponding to the primary key.

        Raises:
            KeyError: If primary key not found in the backend.
        """
        raise NotImplementedError

    # ***

    def get_all(
        self,
        since=None,
        until=None,
        **kwargs
    ):
        """
        Return all facts within a given timeframe that match given search terms.

        # FIXME/2018-06-11: (lb): Update args help... this is stale:

        Args:
            since (datetime.datetime, optional): Consider only Facts
                starting at or after this date. If a time is not specified,
                "00:00:00" is used; otherwise the time of the object is used.
                Defaults to ``None``.

            until (datetime.datetime, optional): Consider only Facts ending
                before or at this date. If not time is specified, "00:00:00"
                is used. Defaults to ``None``.

            filter_term (str, optional): Only consider ``Facts`` with this
                string as part of their associated ``Activity.name``

            deleted (boolean, optional): False to restrict to non-deleted
                Facts; True to find only those marked deleted; None to find
                all.

            order (string, optional): 'asc' or 'desc'; re: Fact.start.

        Returns:
            list: List of ``Facts`` matching given specifications.

        Raises:
            TypeError: If ``since`` or ``until`` are not ``datetime.date``,
                ``datetime.time`` or ``datetime.datetime`` objects.

            ValueError: If ``until`` is before ``since``.

        Note:
            * This public function only provides some sanity checks and
                normalization. The actual backend query is handled by ``_get_all``.
            * ``search_term`` should be prefixable with ``not`` in order to
                invert matching.
            * This does only return proper facts and does not include any
                existing 'ongoing fact'.
            * This method will *NOT* return facts that start before and end
                after (e.g. that span more than) the specified timeframe.
        """
        def _get_all_verify_since_until(since, until, **kwargs):
            self.store.logger.debug(
                _('since: {since} / until: {until}').format(
                    since=since, until=until
                )
            )
            since_dt = _get_all_verify_since(since)
            until_dt = _get_all_verify_until(until)
            if since_dt and until_dt and (until_dt <= since_dt):
                message = _("`until` cannot be earlier than `since`.")
                self.store.logger.debug(message)
                raise ValueError(message)
            return self._get_all(since=since_dt, until=until_dt, **kwargs)

        def _get_all_verify_since(since):
            if since is None:
                return since

            if isinstance(since, datetime.datetime):
                # isinstance(datetime.datetime, datetime.date) returns True,
                # which is why we need to catch this case first.
                since_dt = since
            elif isinstance(since, datetime.date):
                # The user specified a date, but not a time. Assume midnight.
                # MAYBE: Use config['day_start'] and subtract a day minus a minute?
                self.store.logger.debug(
                    _('Using midnight as clock time for `since` date.')
                )
                day_start = self.config['time.day_start']
                since_dt = datetime.datetime.combine(since, day_start)
            elif isinstance(since, datetime.time):
                since_dt = datetime.datetime.combine(datetime.date.today(), since)
            else:
                message = _(
                    'You need to pass either a datetime.date, datetime.time'
                    ' or datetime.datetime object.'
                )
                self.store.logger.debug(message)
                raise TypeError(message)
            return since_dt

        def _get_all_verify_until(until):
            if until is None:
                return until

            if isinstance(until, datetime.datetime):
                # isinstance(datetime.datetime, datetime.date) returns True,
                # which is why we need to except this case first.
                until_dt = until
            elif isinstance(until, datetime.date):
                # MAYBE: (lb): Feels weird that since defaults to midnight,
                #   but until defaults to 'day_start' plus a day...
                #   (need to TESTME to really feel what's going on).
                until_dt = self.day_end_datetime(until)
            elif isinstance(until, datetime.time):
                until_dt = datetime.datetime.combine(datetime.date.today(), until)
            else:
                message = _(
                    'You need to pass either a datetime.date, datetime.time'
                    ' or datetime.datetime object.'
                )
                raise TypeError(message)
            return until_dt

        return _get_all_verify_since_until(since, until, **kwargs)

    # ***

    def _get_all(
        self,
        since=None,
        until=None,
        endless=False,
        partial=False,
        include_usage=True,
        count_results=False,
        # QUESTION: (lb): Would a Fact ever be "hidden"?
        #   (Other than what momentaneous toggles.)
        hidden=False,
        deleted=False,
        key=None,
        search_term='',
        activity=False,
        category=False,
        sort_col='',
        sort_order='',
        limit='',
        offset='',
    ):
        """
        Return a list of ``Facts`` matching given criteria.

        Args:
            start_date (datetime.datetime, optional): Consider only Facts
                starting at or after this datetime. Defaults to ``None``.
            end_date (datetime.datetime): Consider only Facts ending before or at
                this datetime. Defaults to ``None``.
            search_term (str): Cases insensitive strings to match
                ``Activity.name`` or ``Category.name``.
            deleted (boolean, optional): False to restrict to non-deleted
                Facts; True to find only those marked deleted; None to find
                all.
            partial (bool): If ``False`` only facts which start *and* end
                within the timeframe will be considered.
            order (string, optional): 'asc' or 'desc', 'natch.

        Returns:
            list: List of ``Facts`` matching given specifications.

        Note:
            In contrast to the public ``get_all``, this method actually handles the
            backend query.
        """
        raise NotImplementedError

    # ***

    def get_today(self):
        """
        Return all facts for today, while respecting ``day_start``.

        Returns:
            list: List of ``Fact`` instances.

        Note:
            * This does only return proper facts and does not include any
              existing 'ongoing fact'.
        """
        self.store.logger.debug(_("Returning today's facts"))

        today = self.store.now.date()
        since = datetime.datetime.combine(today, self.config['time.day_start'])
        until = self.day_end_datetime(today)
        return self.get_all(since=since, until=until)

    def day_end_datetime(self, end_date=None):
        if end_date is None:
            end_date = self.store.now.date()
        start_time = self.config['time.day_start']
        return fact_time.day_end_datetime(end_date, start_time)

    # ***

    # 2020-01-28: This method only called by nark/tests/. dob replaced its
    # stop-fact mechanism with add_fact, to handle all the possible stop
    # scenarios (e.g., specifying an end time or not; appending the endless
    # fact's description, if it's already got one, etc.).
    def stop_current_fact(self, end_hint=None):
        """
        Stop current 'ongoing fact'.

        Args:
            end_hint (datetime.timedelta or datetime.datetime, optional): Hint to be
                considered when setting ``Fact.end``. If no hint is provided
                ``Fact.end`` will be ``datetime.datetime.now()``. If a ``datetime`` is
                provided, this will be used as ``Fact.end`` value. If a ``timedelta``
                is provided it will be added to ``datetime.datetime.now()``.
                If you want the computed ``end`` to be *before* ``now()``
                you can pass negative ``timedelta`` values. Defaults to None.

        Returns:
            nark.Fact: The stored fact.

        Raises:
            TypeError: If ``end_hint`` is not a ``datetime.datetime`` or
                ``datetime.timedelta`` instance or ``None``.
            ValueError: If there is no currently 'ongoing fact' present.
            ValueError: If the final end value (due to the hint) is before
                the fact's start value.
        """
        self.store.logger.debug(_("Stopping 'ongoing fact'."))

        if not (
            (end_hint is None)
            or isinstance(end_hint, datetime.datetime)
            or isinstance(end_hint, datetime.timedelta)
        ):
            raise TypeError(_(
                "The 'end_hint' you passed needs to be either a"
                "'datetime.datetime' or 'datetime.timedelta' instance."
            ))

        if end_hint:
            if isinstance(end_hint, datetime.datetime):
                end = end_hint
            else:
                end = self.store.now + end_hint
        else:
            end = self.store.now

        fact = self.get_current_fact()
        if fact:
            if fact.start > end:
                raise ValueError(_(
                    'Cannot end the Fact before it started.'
                    ' Try editing the Fact instead.'
                ))
            else:
                fact.end = end
            new_fact = self.save(fact)
            self.store.logger.debug(_("Current fact is now history!"))
        else:
            message = _("Trying to stop a non existing ongoing fact.")
            self.store.logger.debug(message)
            raise ValueError(message)
        return new_fact

    # ***

    def get_current_fact(self):
        """
        Provide a way to retrieve any existing 'ongoing fact'.

        Returns:
            nark.Fact: An instance representing our current
                <ongoing fact>.

        Raises:
            KeyError: If no ongoing fact is present.
        """
        def _get_current_fact():
            self.store.logger.debug(_("Looking for the 'ongoing fact'."))
            # See alternatively:
            #   facts = self.get_all(endless=True)
            facts = self.endless()
            ensure_one_or_fewer_ongoing(facts)
            ensure_one_or_more_ongoing(facts)
            return facts[0]

        def ensure_one_or_fewer_ongoing(facts):
            if len(facts) <= 1:
                return
            msg = '{} IDs: {}'.format(
                _('More than 1 ongoing Fact found. Your database is whacked out!!'),
                ', '.join([str(fact.pk) for fact in facts]),
            )
            self.store.logger.debug(msg)
            raise Exception(msg)

        def ensure_one_or_more_ongoing(facts):
            if facts:
                return
            # See also: NO_ACTIVE_FACT_HELP
            message = _("No ongoing Fact found.")
            self.store.logger.debug(message)
            raise KeyError(message)

        return _get_current_fact()

    # ***

    def find_latest_fact(self, restrict=None):
        assert not restrict or restrict in ['ended', 'ongoing', ]
        fact = None
        if not restrict or restrict == 'ongoing':
            try:
                fact = self.get_current_fact()
            except KeyError:
                fact = None
            except Exception:
                # (lb): Unexpected! This could mean more than one ongoing Fact found!
                raise
        if fact is None and restrict != 'ongoing':
            results = self.get_all(
                sort_col='start', sort_order='desc', limit=1, exclude_ongoing=True,
            )
            if len(results) > 0:
                assert len(results) == 1
                fact, = results
        return fact

    # ***

    def find_oldest_fact(self):
        fact = None
        results = self.facts.get_all(
            sort_col='start', sort_order='asc', limit=1,
        )
        if len(results) > 0:
            assert len(results) == 1
            fact, = results
        return fact

    # ***

    def cancel_current_fact(self, purge=False):
        """
        Delete the current, ongoing, endless Fact. (Really just mark it deleted.)

        Returns:
            None: If everything worked as expected.

        Raises:
            KeyError: If no ongoing fact is present.
        """
        self.store.logger.debug(_("Cancelling 'ongoing fact'."))

        fact = self.get_current_fact()
        if not fact:
            message = _("Trying to stop a non existing ongoing fact.")
            self.store.logger.debug(message)
            raise KeyError(message)
        self.remove(fact, purge)
        return fact

    # ***

    def starting_at(self, fact):
        """
        Return the fact starting at the moment in time indicated by fact.start.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            IntegrityError: If more than one Fact found at given time.
        """
        raise NotImplementedError

    # ***

    def ending_at(self, fact):
        """
        Return the fact ending at the moment in time indicated by fact.end.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            IntegrityError: If more than one Fact found at given time.
        """
        raise NotImplementedError

    # ***

    def antecedent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately preceding the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The antecedent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        raise NotImplementedError

    # ***

    def subsequent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately following the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The subsequent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        raise NotImplementedError

    # ***

    def strictly_during(self, start, end, result_limit=1000):
        """
        Return the fact(s) strictly contained within a start and end time.

        Args:
            start (datetime.datetime):
                Start datetime of facts to find.

            end (datetime.datetime):
                End datetime of facts to find.

            result_limit (int):
                Maximum number of facts to find, else log warning message.

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        raise NotImplementedError

    # ***

    def surrounding(self, fact_time):
        """
        Return the fact(s) at the given moment in time.
        Note that this excludes a fact that starts or ends at this time.
        (See antecedent and subsequent for finding those facts.)

        Args:
            fact_time (datetime.datetime):
                Time of fact(s) to match.

        Returns:
            list: List of ``nark.Facts`` instances.

        Raises:
            IntegrityError: If more than one Fact found at given time.
        """
        raise NotImplementedError

    # ***

    def endless(self):
        """
        Return any facts without a fact.start or fact.end.

        Args:
            <none>

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        raise NotImplementedError




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest
from freezegun import freeze_time
from nark.helpers import fact_time, parse_time
from nark.helpers.parse_errors import ParserInvalidDatetimeException


class TestGetDayEnd(object):
    @pytest.mark.parametrize(('day_start', 'expectation'), [
        (datetime.time(0, 0, 0), datetime.time(23, 59, 59)),
        (datetime.time(5, 30, 0), datetime.time(5, 29, 59)),
        (datetime.time(23, 59, 59), datetime.time(23, 59, 58)),
        (datetime.time(14, 44, 23), datetime.time(14, 44, 22)),
    ])
    def test_various_day_start_times(self, base_config, day_start, expectation):
        """Ensure that resulting end times match our expectation given ``day_start``-"""
        base_config['time.day_start'] = day_start
        assert fact_time.day_end_time(day_start) == expectation


class TestEndDayToDatetime(object):
    @pytest.mark.parametrize(
        ('day_start', 'expectation'),
        [
            (datetime.time(0, 0, 0), datetime.datetime(2015, 4, 15, 23, 59, 59)),
            (datetime.time(5, 30, 0), datetime.datetime(2015, 4, 16, 5, 29, 59)),
            (datetime.time(23, 59, 59), datetime.datetime(2015, 4, 16, 23, 59, 58)),
            (datetime.time(14, 44, 23), datetime.datetime(2015, 4, 16, 14, 44, 22)),
        ],
    )
    def test_same_end_day_various_day_starts(self, day_start, expectation):
        """
        Ensure that resulting ``end datetimes`` match our expectation given ``day_end``
        """
        end_day = datetime.datetime(2015, 4, 15)
        assert fact_time.day_end_datetime(end_day, day_start) == expectation


@freeze_time('2015-12-10 12:30')
class TestParseTime(object):
    @pytest.mark.parametrize(('time', 'expectation'), [
        ('18:55', datetime.datetime(2015, 12, 9, 18, 55)),
        ('18:55:34', datetime.datetime(2015, 12, 9, 18, 55, 34)),
        ('2014-12-10', datetime.datetime(2014, 12, 10, 0, 0)),
        ('2015-10-02 18:12', datetime.datetime(2015, 10, 2, 18, 12)),
        ('2015-10-02 18:12:33', datetime.datetime(2015, 10, 2, 18, 12, 33)),
    ])
    def test_parse_dated_valid_times(self, time, expectation):
        """Make sure that given times are parsed as expected."""
        parsed = parse_time.parse_dated(time, time_now=datetime.datetime.now())
        assert parsed == expectation

    @pytest.mark.parametrize('time', ['18 55', '18:555', '2014 01 04 12:30'])
    def test_parse_dated_invalid_times(self, time):
        """Ensure that invalid times throw an exception."""
        with pytest.raises(ParserInvalidDatetimeException):
            # F841 local variable 'parsed_' is assigned to but never used
            parsed_ = parse_time.parse_dated(  # noqa: F841
                time, time_now=datetime.datetime.now(), cruftless=True,
            )
            assert False  # Unreachable.


class TestValidateStartEndRange(object):
    """Unittests for validation function."""

    @pytest.mark.parametrize('range', (
        (datetime.datetime(2016, 12, 1, 12, 30), datetime.datetime(2016, 12, 1, 12, 45)),
        (datetime.datetime(2016, 1, 1, 12, 30), datetime.datetime(2016, 12, 1, 12, 45)),
        (datetime.datetime(2016, 1, 1, 12, 30), datetime.datetime(2016, 12, 1, 1, 45)),
    ))
    def test_valid_ranges(self, range):
        """Make sure that ranges with end > start pass validation."""
        result = fact_time.must_not_start_after_end(range)
        assert result == range

    @pytest.mark.parametrize('range', (
        (datetime.datetime(2016, 12, 1, 12, 30), datetime.datetime(2016, 12, 1, 10, 45)),
        (datetime.datetime(2016, 1, 13, 12, 30), datetime.datetime(2016, 1, 1, 12, 45)),
    ))
    def test_invalid_ranges(self, range):
        """Make sure that ranges with start > end fail validation."""
        with pytest.raises(ValueError):
            fact_time.must_not_start_after_end(range)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

from gettext import gettext as _

import re
from datetime import timedelta

import lazy_import

from .fact_time import (
    RE_PATTERN_RELATIVE_CLOCK,
    RE_PATTERN_RELATIVE_DELTA,
    datetime_from_clock_prior
)
from .parse_errors import ParserInvalidDatetimeException

# Profiling: load iso8601: ~ 0.004 secs.
iso8601 = lazy_import.lazy_module('iso8601')

__all__ = (
    'HamsterTimeSpec',
    'parse_dated',
    'parse_clock_time',
    'parse_datetime_iso8601',
    'parse_relative_minutes',
)


# =================
# Notes on ISO 8601
# =================
#
# ASIDE: (lb): It's not my intent to write a datetime parser (there are plenty
# out there!), but our flexible factoid format lets the user specify datetimes
# in different, non-standard ways. E.g., the user can use relative time, which
# needs business logic to transform into a real datetime. So it's up to us to
# at least parse the datetime well enough to identify what type of format it's
# in, and then to either process it ourselves, or filter it through ``iso8601``
# or ``dateparser``.
#
# Some examples of ISO 8601 compliant datetimes
# ---------------------------------------------
#
#     ✓ 2018-05-14
#     ✓ 2018-05-14T22:29:24.123456+00:00
#     ✓ 2018-05-14T22:29:24+00:00
#     ✓ 2018-05-14T22:29:24Z
#     ✓ 20180514T222924Z
#
# Not all formats are supported by the ``iso8601`` parser
# -------------------------------------------------------
#
#     ✗ 2018-W20
#     ✗ 2018-W20-1
#     ✗ --05-14
#     ✗ 2018-134
#     ✓ 2018-12
#     ✗ 201805
#     ✓ 2018
#   __ ______________________________________________
#     ^ indicates if parse-worthy by iso8601 (✓ or ✗).
#
# And the ``iso8601`` parser also supports an extended format
# -----------------------------------------------------------
#
#   - The iso8601 parser allows ' ' in lieu of 'T'.
#
# The iso8601 parser format is: ``Date [Time [Timezone]]``
# --------------------------------------------------------
#
#   - Date and time are separated by ' ' or 'T'.
#
#   - Timezone immediately follow Time (no delimiter/space).
#
#   - Dates: YYYY-MM-DD | YYYYMMDD | YYYY-MM | YYYY
#
#   - Times: hh:mm:ss.nn | hhmmss.nn | hh:mm | hhmm | hh
#
#   - Time zones: <nothing> | Z | +/-hh:mm | +/-hhmm | +/-hh
#
#   - You can specify days or months without leading 0s [(lb): but why?].

class HamsterTimeSpec(object):
    """"""
    RE_HAMSTER_TIME = None

    def __init__(self):
        """Not implemented: Use class as static/global, not instantiated."""
        raise NotImplementedError

    @staticmethod
    def discern(hamster_time):
        """
        Check for magic datetimes:
          - '+/-n' relative;
          - 'nn:nn' clocktime;
          - ISO 8601 datetime.

        NOTE: This fcn. does not make datetime.datetime's; that's up to the caller.
        """
        dt, type_dt, sep, rest = None, None, None, None

        if HamsterTimeSpec.RE_HAMSTER_TIME is None:
            HamsterTimeSpec.setup_re()

        match = HamsterTimeSpec.RE_HAMSTER_TIME.match(hamster_time)
        if match is not None:
            say_what = match.groupdict()
            if say_what['relative']:
                assert dt is None
                dt = say_what['relative']
                type_dt = 'relative'
            if say_what['clock_time']:
                assert dt is None
                dt = say_what['clock_time']
                type_dt = 'clock_time'
            if say_what['datetime']:
                assert dt is None
                dt = say_what['datetime']
                type_dt = 'datetime'
            assert dt is not None
            sep = say_what['sep']
            rest = say_what['rest']

        return dt, type_dt, sep, rest

    @staticmethod
    def setup_re():
        # NOTE: This pattern isn't perfect; and that's why we use the
        #       iso8601.parse_date routine.
        #
        #       (lb): It's because we use the ()? optionals.
        #       If one of the optionals is formatted incorrectly,
        #       the pattern here happily ignores it, because ?
        #       For instance, this matches, but the microseconds has an error:
        #
        #           RE_HAMSTER_TIME.match('2018-05-14 22:29:24.123x456+00:02')

        # Never forget! Hamster allows relative time!
        pattern_relative = (
            '(?P<relative>([-+]?(\d+h)|[-+](\d+h)?\d+m?))'
        )

        pattern_just_clock = (
            '(?P<clock_time>\d{1,2}:?\d{2}(:\d{2})?)'
        )

        # (lb): Treat 4 digits as clock time, not year, i.e.,
        #   `2030` should be 10:30 PM, not Jan 01, 2030.
        # This steals colon-less clock times:
        #   '(?:(\d{8}|\d{4}|\d{4}-\d{1,2}(-\d{1,2})?))'
        pattern_date = (
            '(?:(\d{8}|\d{4}-\d{1,2}(-\d{1,2})?))'
        )
        pattern_time = (  # noqa: E131
            # (lb): We could allow 3-digit times... but, no.
            #   '(?:\d{1,2})'
            '(?:\d{2})'
            '(?::?\d{2}'
                '(?::?\d{2}'
                    '(?:\.\d+)?'
                ')?'
            ')?'
        )
        pattern_zone = (  # noqa: E131
            '(?:('
                'Z'
            '|'
                '[+-]\d{2}(:?\d{2})?'
            '))?'
        )
        pattern_datetime = (
            '(?P<datetime>{}([ T]{}{})?)'
            .format(pattern_date, pattern_time, pattern_zone)
        )

        hamster_pattern = (
            '(^|\s)({}|{}|{})(?P<sep>[,:]?)(?=\s|$)(?P<rest>.*)'
            .format(
                pattern_relative,
                pattern_just_clock,
                pattern_datetime,
            )
        )

        # Use re.DOTALL to match newlines, which might be part
        # of the <rest> of the factoid.
        HamsterTimeSpec.RE_HAMSTER_TIME = re.compile(hamster_pattern, re.DOTALL)

    @staticmethod
    def has_time_of_day(raw_dt):
        # Assuming format is year-mo-day separated from time of day by space or 'T'.
        parts = re.split(r' |T', raw_dt)
        if len(parts) != 2:
            return False
        assert re.match(RE_PATTERN_RELATIVE_CLOCK, parts[1]) is not None
        return True


# ***

# FIXME: Use this fcn. to support relative/clock-time --since/--until.
def parse_dated(dated, time_now, cruftless=False):
    """"""
    def _parse_dated():
        if not isinstance(dated, str):
            # Let BaseFactManager.get_all() process, if not already datetime.
            return dated
        dt, type_dt, sep, rest = HamsterTimeSpec.discern(dated)
        if cruftless and rest:
            msg = _('Found more than datetime in')
            plus_sep = sep and ' + ‘{}’'.format(sep) or ''
            raise ParserInvalidDatetimeException(
                '{} “{}”: ‘{}’{} + ‘{}’'
                .format(msg, dated, str(dt), plus_sep, rest)
            )
        if dt is None:
            raise ParserInvalidDatetimeException(
                '{}: “{}”'.format(_('Unparseable datetime'), dated)
            )
        return datetime_from_discerned(dated, dt, type_dt)

    def datetime_from_discerned(dated, dt, type_dt):
        if type_dt == 'datetime':
            # FIXME: (lb): Implement timezone/local_tz.
            dt_suss = parse_datetime_iso8601(dt, must=True, local_tz=None)
        # else, relative time, or clock time; let caller handle.
        elif type_dt == 'clock_time':
            clock_time = parse_clock_time(dt)
            dt_suss = datetime_from_clock_prior(time_now, clock_time)
        else:
            assert type_dt == 'relative'
            rel_mins, negative = parse_relative_minutes(dt)
            dt_suss = time_now + timedelta(minutes=rel_mins)
        return dt_suss

    return _parse_dated()


# ***

def parse_clock_time(clock_time):
    parsed_ct = None
    match = RE_PATTERN_RELATIVE_CLOCK.match(clock_time)
    if match:
        parts = match.groupdict()
        parsed_ct = (
            parts['hours'],
            parts['minutes'],
            parts['seconds'] or '0',
        )
    return parsed_ct


# ***

def parse_datetime_iso8601(datepart, must=False, local_tz=None):
    try:
        # NOTE: Defaults to datetime.timezone.utc.
        #       Uses naive if we set default_timezone=None.
        parsed = iso8601.parse_date(datepart, default_timezone=local_tz)
    except iso8601.iso8601.ParseError as err:
        parsed = None
        if must:
            raise ParserInvalidDatetimeException(_(
                'Unable to parse iso8601 datetime: {} [{}]'
                .format(datepart, str(err))
            ))
    return parsed


# ***

def parse_relative_minutes(rel_time):
    rel_mins = None
    negative = None
    # NOTE: re.match checks for a match only at the beginning of the string.
    match = RE_PATTERN_RELATIVE_DELTA.match(rel_time)
    if match:
        parts = match.groupdict()
        rel_mins = 0
        if parts['minutes']:
            rel_mins += int(parts['minutes'])
        if parts['hours']:
            rel_mins += int(parts['hours']) * 60
        if parts['signage'] == '-':
            negative = True  # Because there's no such thang as "-0".
            rel_mins *= -1
        else:
            negative = False
    return rel_mins, negative




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""
Module to provide generic reporting capabilities for easy adaption by clients.

The basic idea is to provide ``Writer`` classes that take care of the bulk
of the setup upon instantiation so all the client needs to do is to call
``write_report`` with a list of ``FactTuples`` as arguments.
"""

from gettext import gettext as _

import csv
import datetime
import sys
from collections import namedtuple

import lazy_import

# Profiling: load icalendar: ~ 0.008 secs.
icalendar = lazy_import.lazy_module('icalendar')

# SYNC_ME: FactTuple and PlaintextWriter's headers.
FactTuple = namedtuple(
    'FactTuple',
    (
        'start',
        'end',
        'duration',
        'activity',
        'category',
        'description',
        'deleted',
    )
)


class ReportWriter(object):
    def __init__(
        self,
        path,
        datetime_format="%Y-%m-%d %H:%M:%S",
        output_b=False,
    ):
        """
        Initiate new instance and open an output file like object.

        Note:
            If you need added bells and wristels (like heading etc.) this would
            probably the method to extend.

        Args:
            path: File like object to be opened. This is where all output
              will be directed to. datetime_format (str): String specifying how
              datetime information is to be rendered in the output.
        """
        self.datetime_format = datetime_format
        # No matter through what loops we jump, at the end of the day py27
        # ``writerow`` will insist on casting our data to binary str()
        # instances. This clearly conflicts with any generic open() that provides
        # transparent text input/output and would take care of the encoding
        # instead.

        # [FIXME]
        # If it turns out that this is specific to csv handling we may move it
        # there and use a simpler default behaviour for our base method.
        if not path:
            self.file = sys.stdout
        else:
            self.open_file(path, output_b=output_b)

    def open_file(self, path, output_b=False):
        if not output_b:
            self.file = open(path, 'w', encoding='utf-8')
        else:
            self.file = open(path, 'wb')

    def write_report(self, facts):
        """
        Write facts to output file and close the file like object.

        Args:
            facts (Iterable): Iterable of ``nark.Fact`` instances to export.

        Returns:
            None: If everything worked as expected.
        """
        for fact in facts:
            self._write_fact(self._fact_to_tuple(fact))
        self._close()

    def _fact_to_tuple(self, fact):
        """
        Convert a ``Fact`` to its normalized tuple.

        This is where all type conversion for ``Fact`` attributes to strings as well
        as any normalization happens.

        Note:
            Because different writers may require different types, we need to so this
            individualy.

        Args:
            fact (nark.Fact): Fact to be converted.

        Returns:
            FactTuple: Tuple representing the original ``Fact``.
        """
        raise NotImplementedError

    def _write_fact(self, fact):
        """
        Represent one ``Fact`` in the output file.

        What this means exactly depends on the format and kind of output.
        At this point all type conversions and normalization have already been done.

        Args:
            fact (FactTuple): The individual fact to be written.

        Returns:
            None
        """
        raise NotImplementedError

    def _close(self):
        """Default teardown method."""
        if self.file is sys.stdout:
            return
        self.file.close()


class PlaintextWriter(ReportWriter):
    # HINT: For list of dialects:
    #   >>> import csv
    #   >>> csv.list_dialects()
    #   ['excel-tab', 'excel', 'unix']
    def __init__(
        self,
        path,
        duration_fmt,
        datetime_format="%Y-%m-%d %H:%M:%S",
        output_b=False,
        dialect='excel',
        **fmtparams
    ):
        """
        Initialize a new instance.

        Besides our default behaviour we create a localized heading.
        Also, we need to make sure that our heading is UTF-8 encoded on python 2!
        In that case ``self.file`` will be openend in binary mode and ready to accept
        those encoded headings.
        """
        super(PlaintextWriter, self).__init__(
            path, datetime_format, output_b=output_b,
        )
        self.csv_writer = csv.writer(self.file, dialect=dialect, **fmtparams)
        # SYNC_ME: FactTuple and PlaintextWriter's headers.
        headers = (
            _("start time"),
            _("end time"),
            _("duration minutes"),
            _("activity"),
            _("category"),
            _("description"),
            _("deleted"),
        )
        results = []
        for header in headers:
            results.append(header)
        self.csv_writer.writerow(results)
        self.duration_fmt = duration_fmt

    def _fact_to_tuple(self, fact):
        """
        Convert a ``Fact`` to its normalized tuple.

        This is where all type conversion for ``Fact`` attributes to strings as well
        as any normalization happens.

        Args:
            fact (nark.Fact): Fact to be converted.

        Returns:
            FactTuple: Tuple representing the original ``Fact``.
        """
        # Fields that allow ``None`` values will be represented by empty ''s.
        # FIXME/DRY/2020-01-16: (lb): This block repeated throughout this file:
        if fact.activity:
            activity = fact.activity.name
        else:
            activity = ''
        if fact.category:
            category = fact.category.name
        else:
            category = ''
        description = fact.description or ''

        start = fact.start.strftime(self.datetime_format) if fact.start else ''
        end = fact.end.strftime(self.datetime_format) if fact.end else ''

        return FactTuple(
            start=start,
            end=end,
            duration=fact.format_delta(style=self.duration_fmt),
            activity=activity,
            category=category,
            description=description,
            deleted=str(fact.deleted),
        )

    def _write_fact(self, fact_tuple):
        """
        Write a single fact.

        On python 2 we need to make sure we encode our data accordingly so we
        can feed it to our file object which in this case needs to be opened in
        binary mode.
        """
        results = []
        for value in fact_tuple:
            results.append(value)
        self.csv_writer.writerow(results)


class CSVWriter(PlaintextWriter):
    def __init__(self, path, datetime_format="%Y-%m-%d %H:%M:%S"):
        super(CSVWriter, self).__init__(
            path,
            # (lb): I figured using 'excel' dialect would be enough,
            #   but scientificsteve/mr_custom does it different... and
            #   I did not test dialect='excel'
            # MAYBE: (lb): Test dialect='excel' without remaining params.
            #   Or not. Depends how much you care about robustness in the
            #   CLI, or if you just want the dob-start command to work
            #   (that's all I'm really doing here! Except the perfectionist
            #   in me also wanted to make all tests work and to see how much
            #   coverage there is -- and I'm impressed! Project Hamster is so
            #   very well covered, it's laudatory!).
            #
            #  dialect='excel',
            #
            # EXPLAIN/2018-05-05: (lb): What did scientificsteve use '%M'
            #   and not '%H:%M'?
            duration_fmt='%M',
            datetime_format=datetime_format,
            # EXPLAIN/2018-05-05: (lb): ',' is also the default delimiter.
            #   How if this different than the default dialect='excel'?
            #   It's probably not....
            delimiter=str(','),
            quoting=csv.QUOTE_MINIMAL,
        )


class TSVWriter(PlaintextWriter):
    def __init__(self, path, datetime_format="%Y-%m-%d %H:%M:%S"):
        super(TSVWriter, self).__init__(
            path,
            duration_fmt='%H:%M',
            datetime_format=datetime_format,
            dialect='excel-tab',
        )


class ICALWriter(ReportWriter):
    """A simple ical writer for fact export."""
    def __init__(self, path, datetime_format="%Y-%m-%d %H:%M:%S"):
        """
        Initiate new instance and open an output file like object.

        Args:
            path: File like object to be opend. This is where all output
                will be directed to. datetime_format (str): String specifying
                how datetime information is to be rendered in the output.
        """
        super(ICALWriter, self).__init__(path, datetime_format, output_b=True)
        self.calendar = icalendar.Calendar()

    def _fact_to_tuple(self, fact):
        """
        Convert a ``Fact`` to its normalized tuple.

        This is where all type conversion for ``Fact`` attributes to strings as
            well as any normalization happens.

        Note:
            Because different writers may require different types, we need to
                so this individualy.

        Args:
            fact (nark.Fact): Fact to be converted.

        Returns:
            FactTuple: Tuple representing the original ``Fact``.
        """
        # Fields that allow ``None`` values will be represented by empty ''s.
        # FIXME/DRY/2020-01-16: (lb): This block repeated throughout this file:
        if fact.activity:
            activity = fact.activity.name
        else:
            activity = ''
        if fact.category:
            category = fact.category.name
        else:
            category = ''
        description = fact.description or ''

        return FactTuple(
            start=fact.start,
            end=fact.end,
            duration=None,
            activity=activity,
            category=category,
            description=description,
            deleted=str(fact.deleted),
        )

    def _write_fact(self, fact_tuple):
        """
        Write a singular fact to our report.

        Note:
            * ``dtent`` is non-inclusive according to Page 54 of RFC 5545

        Returns:
            None: If everything worked out alright.
        """
        # [FIXME]
        # It apears that date/time requirements for VEVENT have changed between
        # RFCs. 5545 now seems to require a 'dstamp' and a 'uid'!
        event = icalendar.Event()
        event.add('dtstart', fact_tuple.start)
        event.add('dtend', fact_tuple.end + datetime.timedelta(seconds=1))
        event.add('categories', fact_tuple.category)
        event.add('summary', fact_tuple.activity)
        event.add('description', fact_tuple.description)
        self.calendar.add_component(event)

    def _close(self):
        """Custom close method to make sure the calendar is actually writen do disk."""
        self.file.write(self.calendar.to_ical())
        return super(ICALWriter, self)._close()


class XMLWriter(ReportWriter):
    """Writer for a basic xml export."""
    # (lb): @elbenfreund noted that XMLWriter copied from 'legacy hamster':
    #   Authored by tstriker <https://github.com/tstriker>. Docstrings by elbenfreund.
    #   https://github.com/projecthamster/hamster/blame/66ed9270c6f0070a4548aca9f070517cc13c85ae
    #       /src/hamster/reports.py#L159
    #   (Other than this class, the nark code authors are either:
    #    landonb (2018-2020); or elbenfreund (2015-2017).)

    def __init__(self, path, datetime_format="%Y-%m-%d %H:%M:%S"):
        """Setup the writer including a main xml document."""
        super(XMLWriter, self).__init__(path, datetime_format, output_b=True)
        # Profiling: load Document: ~ 0.004 secs.
        from xml.dom.minidom import Document
        self.document = Document()
        self.fact_list = self.document.createElement("facts")

    def _fact_to_tuple(self, fact):
        """
        Convert a ``Fact`` to its normalized tuple.

        This is where all type conversion for ``Fact`` attributes to strings as
        well as any normalization happens.

        Note:
            Because different writers may require different types, we need to
            do this individually.

        Args:
            fact (nark.Fact): Fact to be converted.

        Returns:
            FactTuple: Tuple representing the original ``Fact``.
        """
        # Fields that allow ``None`` values will be represented by empty ''s.
        # FIXME/DRY/2020-01-16: (lb): This block repeated throughout this file:
        if fact.activity:
            activity = fact.activity.name
        else:
            activity = ''
        if fact.category:
            category = fact.category.name
        else:
            category = ''
        description = fact.description or ''

        return FactTuple(
            start=fact.start.strftime(self.datetime_format),
            end=fact.end.strftime(self.datetime_format),
            duration=fact.format_delta(style='%M'),
            activity=activity,
            category=category,
            description=description,
            deleted=str(fact.deleted),
        )

    def _write_fact(self, fact_tuple):
        """
        Create new fact element and populate attributes.

        Once the child is prepared append it to ``fact_list``.
        """
        fact = self.document.createElement("fact")

        # MAYBE/2018-04-22: (lb): Should this be start, or start_time? end, or end_time?
        fact.setAttribute('start', fact_tuple.start)
        fact.setAttribute('end', fact_tuple.end)
        fact.setAttribute('activity', fact_tuple.activity)
        fact.setAttribute('duration', fact_tuple.duration)
        fact.setAttribute('category', fact_tuple.category)
        fact.setAttribute('description', fact_tuple.description)
        self.fact_list.appendChild(fact)

    def _close(self):
        """
        Append the xml fact list to the main document write file and cleanup.

        ``toxml`` should take care of encoding everything with UTF-8.
        """

        self.document.appendChild(self.fact_list)
        self.file.write(self.document.toxml(encoding='utf-8'))
        return super(XMLWriter, self)._close()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

__all__ = (
    'resolve_attr_or_method',
)


def resolve_attr_or_method(self, prop, **kwargs):
    self_val = getattr(self, prop)
    if callable(self_val):
        self_val = self_val(**kwargs)
    return self_val

